from __future__ import annotations

from typing import TYPE_CHECKING

from scc.constants import HapticPos

if TYPE_CHECKING:
	from scc.mapper import Mapper
import logging
import time

log = logging.getLogger("SCController")

next_id: int = 1  # Used with fallback controller id generator


class Controller:
	"""Base class for all controller drivers. Implementations are in scc.drivers package.

	Derived class should implement every method from here.
	"""

	flags = 0

	def __init__(self) -> None:
		global next_id
		self.mapper: Mapper | None = None
		self._id: int | str = next_id
		next_id += 1
		self.lastTime: float = time.time()
		self.time_elapsed: float = 0.0

	def get_type(self) -> str:
		"""Has to return type identifier

		Returns a short string without spaces
		that describes type of controller which should be unique for each
		driver.
		String is used by UI to assign icons and, along with ID,
		to store controller settings.

		This method has to be overriden.
		"""
		raise RuntimeError("Controller.get_type not overriden")

	def get_id(self) -> int | str:
		"""Returns identifier that has to be unique at least until daemon is restarted.

		Ideally derived from HW device serial number.
		"""
		return self._id

	def get_gui_config_file(self) -> str | None:
		"""Returns file name of json file that GUI can use to load more data about controller

		(background image, button images, available buttons and axes, etc...)
		File name may be absolute path or just name of file in /usr/share/scc

		Returns None if there is no configuration file (GUI will use defaults in such case)
		"""
		return

	def set_mapper(self, mapper: Mapper) -> None:
		"""Sets mapper for controller"""
		self.mapper = mapper

	def get_mapper(self) -> Mapper | None:
		"""Returns mapper set for controller"""
		return self.mapper

	def apply_config(self, config) -> None:
		"""Called from daemon to apply controller configuration stored in config file.

		Does nothing by default.
		"""

	def set_led_level(self, level) -> None:
		"""Configures LED intensity, if supported.

		'level' goes from 0.0 to 100.0
		"""

	def set_gyro_enabled(self, enabled) -> None:
		"""Enables or disables gyroscope, if supported"""

	def get_gyro_enabled(self) -> bool:
		"""Returns True if gyroscope is enabled"""
		return False

	def feedback(self, data) -> None:
		"""Generates feedback effect, if supported.

		'data' is HapticData instance.
		"""

	def turnoff(self) -> None:
		"""Turns off controller, if supported"""

	def disconnected(self) -> None:
		"""Called from daemon after controller is disconnected"""


class HapticData:
	"""Simple container to hold haptic feedback settings"""

	def __init__(self, position, amplitude=512, frequency=4, period=1024, count=1):
		"""'frequency' is used only when emulating touchpad

		and describes how many pixels should mouse travel between two feedback ticks.
		"""
		data = tuple([int(x) for x in (position, amplitude, period, count)])
		if data[0] not in (HapticPos.LEFT, HapticPos.RIGHT, HapticPos.BOTH):
			raise ValueError("Invalid position")
		for i in (1, 2, 3):
			if data[i] > 0x8000 or data[i] < 0:
				raise ValueError("Value out of range: %s", data[i])
		# frequency is multiplied by 1000 just so I don't have big numbers everywhere;
		# it's float until here, so user still can make pad squeak if he wish
		frequency = int(max(1.0, frequency * 1000.0))

		self.data = data  # send to controller
		self.frequency = frequency  # used internally

	def with_position(self, position) -> HapticData:
		"""Creates copy of HapticData with position value changed"""
		trash, amplitude, period, count = self.data
		return HapticData(position, amplitude, self.frequency, period, count)

	def get_position(self) -> HapticPos:
		return HapticPos(self.data[0])

	def get_amplitude(self):
		return self.data[1]

	def get_frequency(self) -> float:
		return float(self.frequency) / 1000.0

	def get_period(self):
		return self.data[2]

	def get_count(self):
		return self.data[3]

	def __mul__(self, by) -> HapticData:
		"""Allows multiplying HapticData by scalar to get same values with increased amplitude."""
		position, amplitude, period, count = self.data
		amplitude = min(amplitude * by, 0x8000)
		return HapticData(position, amplitude, self.frequency, period, count)



"""SC Controller - Action Editor - Gesture Component

Handles gesture recognition settings.
"""

import logging

from gi.repository import Gdk, GdkX11, GObject

from scc.actions import Action, NoAction
from scc.gui.action_editor import ActionEditor
from scc.gui.ae import AEComponent
from scc.modifiers import NameModifier
from scc.osd.gesture_display import GestureDisplay
from scc.special_actions import GesturesAction, OSDAction
from scc.tools import _, strip_gesture

log = logging.getLogger("AE.PerAxis")

__all__ = ["GestureComponent"]


class GestureComponent(AEComponent):
	GLADE = "ae/gesture.glade"
	NAME = "gesture"
	CTXS = Action.AC_STICK | Action.AC_PAD
	PRIORITY = 1

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		self.on_wayland = False
		self._edited_gesture = None
		self._grabber = None

	def load(self):
		if AEComponent.load(self):
			# Unlike mose region, gesutres kinda work with XWayland
			self.on_wayland = not isinstance(Gdk.Display.get_default(), GdkX11.X11Display)
			if self.on_wayland:
				self.builder.get_object("lblGestureMessage").set_text(
					_("Note: Gestures are not available with Wayland-based display server"),
				)
				self.builder.get_object("lblGestureMessage").set_visible(True)
				self.builder.get_object("gesture").set_sensitive(False)
			else:
				self._grabber = GestureGrabber(self.editor, self.builder)
			return True

	def set_action(self, mode, action):
		lstGestures = self.builder.get_object("lstGestures")
		lstGestures.clear()
		if isinstance(action, GesturesAction):
			for gstr in action.gestures:
				self._add_gesture(gstr, action.gestures[gstr])

	def _add_gesture(self, gstr, action, select=False):
		lstGestures = self.builder.get_object("lstGestures")
		o = GObject.GObject()
		o.gstr = gstr
		o.action = action
		iter = lstGestures.append((GestureComponent.nice_gstr(gstr), action.describe(Action.AC_MENU), o))
		if select:
			tvGestures = self.builder.get_object("tvGestures")
			tvGestures.get_selection().select_iter(iter)
			self.on_tvGestures_cursor_changed()
			self.on_btEditAction_clicked()

	ARROWS = {
		"U": "↑",
		"D": "↓",
		"L": "←",
		"R": "→",
	}

	@staticmethod
	def nice_gstr(gstr):
		"""Replaces characters UDLR in gesture string with unicode arrows.
		← → ↑ ↓
		"""
		if "i" in gstr:
			gstr = strip_gesture(gstr)
		l = lambda x: GestureComponent.ARROWS[x] if x in GestureComponent.ARROWS else ""
		return "".join(map(l, gstr))

	def get_button_title(self):
		return _("Gestures")

	def handles(self, mode, action):
		return isinstance(action, GesturesAction)

	def on_tvGestures_cursor_changed(self, *a):
		tvGestures = self.builder.get_object("tvGestures")
		btEditAction = self.builder.get_object("btEditAction")
		btRemove = self.builder.get_object("btRemove")
		model, iter = tvGestures.get_selection().get_selected()
		if iter is None:
			btEditAction.set_sensitive(False)
			btRemove.set_sensitive(False)
		else:
			btEditAction.set_sensitive(True)
			btRemove.set_sensitive(True)

	def on_btEditAction_clicked(self, *a):
		"""Handler for "Edit Action" button"""
		tvGestures = self.builder.get_object("tvGestures")
		txGesture = self.builder.get_object("txGesture")
		cbIgnoreStroke = self.builder.get_object("cbIgnoreStroke")
		gesture_editor = self.builder.get_object("gesture_editor")

		model, iter = tvGestures.get_selection().get_selected()
		item = model.get_value(iter, 2)
		self._edited_gesture = item.gstr
		cbIgnoreStroke.set_active("i" in self._edited_gesture)
		txGesture.set_text(GestureComponent.nice_gstr(self._edited_gesture))
		# Setup editor
		e = ActionEditor(self.app, self.on_action_chosen)
		e.set_title(_("Edit Gesture Action"))
		e.set_input("ID", item.action, mode=Action.AC_BUTTON)
		e.add_widget(_("Gesture"), gesture_editor)
		e.hide_modeshift()
		# Display editor
		e.show(self.editor.window)

	def on_btChangeGesture_clicked(self, *a):
		txGesture = self.builder.get_object("txGesture")

		def grabbed(gesture):
			self._edited_gesture = gesture
			txGesture.set_text(GestureComponent.nice_gstr(self._edited_gesture))

		self._grabber.grab(grabbed)

	def on_cbIgnoreStroke_toggled(self, cb):
		txGesture = self.builder.get_object("txGesture")
		if cb.get_active() and "i" not in self._edited_gesture:
			self._edited_gesture = "i" + self._edited_gesture
			txGesture.set_text(GestureComponent.nice_gstr(self._edited_gesture))
		elif not cb.get_active() and "i" in self._edited_gesture:
			self._edited_gesture = self._edited_gesture.strip("i")
			txGesture.set_text(GestureComponent.nice_gstr(self._edited_gesture))

	def on_btRemove_clicked(self, *a):
		tvGestures = self.builder.get_object("tvGestures")
		model, iter = tvGestures.get_selection().get_selected()
		model.remove(iter)
		self.on_tvGestures_cursor_changed()

	def on_btAdd_clicked(self, *a):
		def grabbed(gesture):
			self._add_gesture(gesture, NoAction(), True)

		tvGestures = self.builder.get_object("tvGestures")
		if len(tvGestures.get_model()) == 0:
			# I believe user will not actually find this option, so OSD checkbox
			# is automatically enabled when first item is added
			self.editor.set_osd_enabled(True)
		self._grabber.grab(grabbed)

	def on_sclPrecision_format_value(self, scl, value):
		return "%s%%" % (int(value * 100.0),)

	def on_sclPrecision_value_changed(self, *a):
		self.update()

	def on_btClearTolerance_clicked(self, *a):
		self.builder.get_object("sclPrecision").set_value(GesturesAction.DEFAULT_PRECISION)

	def on_action_chosen(self, id, action, mark_changed=True):
		tvGestures = self.builder.get_object("tvGestures")
		model, iter = tvGestures.get_selection().get_selected()
		item = model.get_value(iter, 2)
		item.gstr = self._edited_gesture
		item.action = action
		model.set_value(iter, 0, GestureComponent.nice_gstr(item.gstr))
		model.set_value(iter, 1, action.describe(Action.AC_MENU))
		self.update()

	def update(self):
		a = GesturesAction()
		tvGestures = self.builder.get_object("tvGestures")
		model, iter = tvGestures.get_selection().get_selected()
		for row in model:
			item = row[2]
			a.gestures[item.gstr] = item.action
			if item.action.name:
				a.gestures[item.gstr] = NameModifier(item.action.name, item.action)
		a.precision = self.builder.get_object("sclPrecision").get_value()
		a = OSDAction(a)
		self.editor.set_action(a)


class GestureGrabber:
	def __init__(self, editor, builder):
		self.editor = editor
		self.builder = builder
		self._callback = None
		self._gd = None
		self._signals = None
		self._gesture = None
		self._repeats = 0
		self.gesture_grabber = self.builder.get_object("gesture_grabber")
		self.txGestureGrab = self.builder.get_object("txGestureGrab")
		self.lblGestureGrabberTitle = self.builder.get_object("lblGestureGrabberTitle")
		self.lblGestureStatus = self.builder.get_object("lblGestureStatus")
		self.rvGestureGrab = self.builder.get_object("rvGestureGrab")
		# Can't use autoconnect for this :(
		self.gesture_grabber.connect("delete-event", self.close)
		self.gesture_grabber.connect("destroy", self.close)
		self.builder.get_object("btnStartGestureOver").connect("clicked", self.start_over)
		self.builder.get_object("btnConfirmGesutre").connect("clicked", self.use)

	def fail(self, *a):
		"""Called when something goes bad, usually because there is
		no controller connected.
		"""
		log.error("Failed to grab gesture: %s", a)

	def disconnect_signals(self):
		"""Disconnects redundant signal handlers.
		Currently only one created in lock_buttons.
		"""
		if self._signals:
			for source, eid in self._signals:
				source.disconnect(eid)
			self._signals = []

	def lock_buttons(self):
		self.disconnect_signals()
		try:
			c = self.editor.app.dm.get_controllers()[0]
			c.lock(
				lambda *a: True,  # success_cb
				self.fail,  # error_cb
				"A",
				"Y",
			)
			self._signals = [(c, c.connect("event", self.on_event))]
		except IndexError:
			# No controllers
			self.fail()

	def on_event(self, c, button, data):
		if self.rvGestureGrab.get_reveal_child():
			if button == "A" and data[0] == 0:
				self.use()
			elif button == "Y" and data[0] == 0:
				self.start_over()

	def grab(self, callback):
		self._callback = callback
		self.start_over()
		self.gesture_grabber.set_transient_for(self.editor.window)
		self.gesture_grabber.show()
		self._create_gd()

	def use(self, *a):
		self._callback(self._gesture)
		self.close()

	def close(self, *a):
		if self._gd:
			self._gd.quit()
		self._gd = None
		self.gesture_grabber.hide()
		return True

	def start_over(self, *a):
		if self.editor.get_id() == "RPAD":
			self.lblGestureGrabberTitle.set_text(_("Draw gesture on RIGHT pad..."))
		elif self.editor.get_id() == "CPAD":
			self.lblGestureGrabberTitle.set_text(_("Draw gesture on Touchpad..."))
		else:
			self.lblGestureGrabberTitle.set_text(_("Draw gesture on LEFT pad..."))
		self.lblGestureStatus.set_label("")
		self.txGestureGrab.set_text("")
		self.rvGestureGrab.set_reveal_child(False)
		self._gesture = None
		self._repeats = 0

	def _create_gd(self):
		"""Creates GestureDisplay object"""
		if self._gd:
			self._gd.quit()
		self._gd = GestureDisplay(self.editor.app.config)
		if self.editor.get_id() == "RPAD":
			self._gd.parse_argumets(["GestureDisplay", "--control-with", "RIGHT"])
		elif self.editor.get_id() == "CPAD":
			self._gd.parse_argumets(["GestureDisplay", "--control-with", "CPAD"])
		else:
			self._gd.parse_argumets(["GestureDisplay", "--control-with", "LEFT"])
		self._gd.use_daemon(self.editor.app.dm)
		self._gd.show()
		self._gd.connect("gesture-updated", self.on_gesture_updated)
		self._gd.connect("destroy", self.on_gesture_recognized)
		self.lock_buttons()

	def on_gesture_updated(self, gd, gstr):
		txt = GestureComponent.nice_gstr(gstr)
		self.txGestureGrab.set_text(txt)
		self.txGestureGrab.set_position(len(txt))

	def on_gesture_recognized(self, gd):
		self.disconnect_signals()
		if gd.get_exit_code() != 0:
			# Canceled or cannot grab controller
			return
		if gd.get_gesture():
			self.on_gesture_updated(gd, gd.get_gesture())
			if self._gesture == None:
				self.lblGestureGrabberTitle.set_text(_("Repeat same gesture or press A button to confirm..."))
				self.rvGestureGrab.set_reveal_child(True)
				self._gesture = gd.get_gesture()
			elif self._gesture == gd.get_gesture():
				self._repeats += 1
				self.lblGestureStatus.set_label(_("Repeated %s times") % (self._repeats,))
			else:
				self.lblGestureStatus.set_label(_("Gesture differs"))

		self._create_gd()



"""ae - Action Editor components"""

import logging
import os

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, GLib, Gtk

from scc.actions import Action, NoAction, XYAction
from scc.gui.editor import ComboSetter
from scc.tools import _, ensure_size

log = logging.getLogger("AE")


class AEComponent(ComboSetter):
	GLADE = None
	NAME = None
	PRIORITY = 0
	# Bit mask of contexes (Action.AC_BUTTON | Action.AC_TRIGGER...) that this
	# compoment can handle.
	CTXS = 0

	def __init__(self, app, editor):
		self.app = app
		self.editor = editor
		self.loaded = False

	def get_button_title(self):
		raise Exception("Implement me!")

	# TODO: Rename this to on_shown
	def shown(self):
		"""Called after user switches TO page"""

	# TODO: Rename this to on_shown
	def hidden(self):
		"""Called after user switches AWAY from page"""

	def on_ok(self, action):
		"""Called when user presses OK, after action is send to main window
		"""

	def load(self):
		"""Performs whatever component needs to get loaded.
		Can be called multiple times without breaking anything, but returns
		True when called first time and then False every to signalize repeated
		call.
		"""
		if self.loaded:
			return False
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(self.app.gladepath, self.GLADE))
		self.widget = self.builder.get_object(self.NAME)
		self.builder.connect_signals(self)
		self.loaded = True
		return True

	def is_loaded(self):
		return self.loaded

	def handles(self, mode, action):
		"""Returns True if component can display and edit specified action.
		If more than one component returns True from 'handles',
		higher PRIORITY is used
		"""
		return False

	def set_action(self, mode, action):
		"""Setups component widgets to display currently set action.
		"""

	def modifier_updated(self):
		"""Called when values of any modifier is changed.
		"""

	def get_widget(self):
		return self.widget


def describe_action(mode, cls, v):
	"""Returns action description with 'v' as parameter, unless unless v is None.
	Returns "not set" if v is None
	"""
	if v is None or type(v) in (
		int,
		float,
		str,
	):
		return _("(not set)")
	if isinstance(v, Action):
		if not mode:
			dsc = v.describe(Action.AC_STICK if cls == XYAction else Action.AC_BUTTON)
		else:
			dsc = v.describe(mode)

		if "\n" in dsc:
			dsc = "<small>" + "\n".join(dsc.split("\n")[0:2]) + "</small>"
		return dsc
	return (cls(v)).describe(mode)



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Axis Component

Assigns emulated axis to trigger
"""

import logging

from scc.actions import Action, NoAction, ResetGyroAction
from scc.gui.ae import AEComponent
from scc.gui.ae.menu_action import MenuActionCofC
from scc.special_actions import (
	ChangeProfileAction,
	ClearOSDAction,
	KeyboardAction,
	OSDAction,
	ShellCommandAction,
	TurnOffAction,
)
from scc.tools import _

log = logging.getLogger("AE.SA")

__all__ = ["SpecialActionComponent"]


class SpecialActionComponent(AEComponent, MenuActionCofC):
	GLADE = "ae/special_action.glade"
	NAME = "special_action"
	CTXS = Action.AC_BUTTON | Action.AC_MENU
	PRIORITY = 0

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		MenuActionCofC.__init__(self)
		self._userdata_load_started = False
		self._recursing = False
		self._current_profile = None

	def load(self):
		if self.loaded:
			return
		AEComponent.load(self)
		cbConfirmWith = self.builder.get_object("cbConfirmWith")
		cbCancelWith = self.builder.get_object("cbCancelWith")
		cbConfirmWith.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) == "-")
		cbCancelWith.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) == "-")

	def shown(self):
		if not self._userdata_load_started:
			self._userdata_load_started = True
			self.load_profile_list()
			self.load_menu_list()

	def confirm_with_same_active(self):
		cbMenuAutoConfirm = self.builder.get_object("cbMenuAutoConfirm")
		return cbMenuAutoConfirm.get_active()

	def set_action(self, mode, action):
		if self.handles(mode, action):
			cb = self.builder.get_object("cbActionType")
			if isinstance(action, TurnOffAction):
				self.set_cb(cb, "turnoff")
			elif isinstance(action, ShellCommandAction):
				self.set_cb(cb, "shell")
				enCommand = self.builder.get_object("enCommand")
				enCommand.set_text(action.command.encode("utf-8"))
			elif isinstance(action, ResetGyroAction):
				self.set_cb(cb, "resetgyro")
			elif isinstance(action, ChangeProfileAction):
				self._current_profile = action.profile
				self.set_cb(cb, "profile")
			elif MenuActionCofC.handles(self, None, action):
				self.set_cb(cb, "menu")
				self.load_menu_data(action)
			elif isinstance(action, KeyboardAction):
				self.set_cb(cb, "keyboard")
			elif isinstance(action, OSDAction):
				self.set_cb(cb, "osd")
				sclOSDTimeout = self.builder.get_object("sclOSDTimeout")
				enOSDText = self.builder.get_object("enOSDText")
				cbOSDSize = self.builder.get_object("cbOSDSize")
				self._recursing = True
				sclOSDTimeout.set_value(action.timeout or 60.1)
				enOSDText.set_text(action.text)
				self.set_cb(cbOSDSize, action.size)
				self._recursing = False
			elif isinstance(action, ClearOSDAction):
				self.set_cb(cb, "clearosd")
			else:
				self.set_cb(cb, "none")

	def on_profiles_loaded(self, profiles):
		cb = self.builder.get_object("cbProfile")
		model = cb.get_model()
		model.clear()
		i, current_index = 0, 0
		for f in sorted(profiles, key=lambda f: f.get_basename()):
			name = f.get_basename()
			if name.endswith(".mod"):
				continue
			if name.startswith("."):
				continue
			if name.endswith(".sccprofile"):
				name = name[0:-11]
			if name == self._current_profile:
				current_index = i
			model.append((name, f, None))
			i += 1

		self._recursing = True
		cb.set_active(current_index)
		self._recursing = False

	def get_button_title(self):
		return _("Special Action")

	def handles(self, mode, action):
		if MenuActionCofC.handles(self, mode, action):
			return True
		if isinstance(action, OSDAction) and action.action is None:
			return True
		return isinstance(
			action,
			(
				NoAction,
				TurnOffAction,
				ShellCommandAction,
				ChangeProfileAction,
				KeyboardAction,
				ClearOSDAction,
				ResetGyroAction,
			),
		)

	def on_cbActionType_changed(self, *a):
		cbActionType = self.builder.get_object("cbActionType")
		stActionData = self.builder.get_object("stActionData")
		key = cbActionType.get_model().get_value(cbActionType.get_active_iter(), 0)
		if key == "shell":
			stActionData.set_visible_child(self.builder.get_object("vbShell"))
			self.on_enCommand_changed()
		elif key == "profile":
			stActionData.set_visible_child(self.builder.get_object("vbProfile"))
			self.on_cbProfile_changed()
		elif key == "keyboard":
			stActionData.set_visible_child(self.builder.get_object("nothing"))
			if not self._recursing:
				self.editor.set_action(KeyboardAction())
		elif key == "clearosd":
			stActionData.set_visible_child(self.builder.get_object("nothing"))
			if not self._recursing:
				self.editor.set_action(ClearOSDAction())
		elif key == "resetgyro":
			stActionData.set_visible_child(self.builder.get_object("nothing"))
			if not self._recursing:
				self.editor.set_action(ResetGyroAction())
		elif key == "osd":
			stActionData.set_visible_child(self.builder.get_object("vbOSD"))
			if not self._recursing:
				self.editor.set_action(OSDAction(""))
		elif key == "menu":
			stActionData.set_visible_child(self.builder.get_object("grMenu"))
			self.on_cbMenus_changed()
		elif key == "turnoff":
			stActionData.set_visible_child(self.builder.get_object("nothing"))
			if not self._recursing:
				self.editor.set_action(TurnOffAction())
		else:  # none
			stActionData.set_visible_child(self.builder.get_object("nothing"))
			if not self._recursing:
				self.editor.set_action(NoAction())

	def on_cbProfile_changed(self, *a):
		"""Called when user chooses profile in selection combo"""
		if self._recursing:
			return
		cb = self.builder.get_object("cbProfile")
		model = cb.get_model()
		iter = cb.get_active_iter()
		if iter is None:
			# Empty list
			return
		f = model.get_value(iter, 1)
		name = f.get_basename()
		if name.endswith(".sccprofile"):
			name = name[0:-11]
		self.editor.set_action(ChangeProfileAction(name))

	def on_enCommand_changed(self, *a):
		if self._recursing:
			return
		enCommand = self.builder.get_object("enCommand")
		self.editor.set_action(ShellCommandAction(enCommand.get_text()))

	def on_osd_settings_changed(self, *a):
		if self._recursing:
			return
		enOSDText = self.builder.get_object("enOSDText")
		sclOSDTimeout = self.builder.get_object("sclOSDTimeout")
		cbOSDSize = self.builder.get_object("cbOSDSize")
		timeout = sclOSDTimeout.get_value()
		size = cbOSDSize.get_model().get_value(cbOSDSize.get_active_iter(), 0)
		self.editor.set_action(OSDAction(0 if timeout > 60.0 else timeout, size, enOSDText.get_text()))

	def on_exMenuControl_activate(self, ex, *a):
		rvMenuControl = self.builder.get_object("rvMenuControl")
		rvMenuControl.set_reveal_child(not ex.get_expanded())

	def on_exMenuPosition_activate(self, ex, *a):
		rvMenuPosition = self.builder.get_object("rvMenuPosition")
		rvMenuPosition.set_reveal_child(not ex.get_expanded())

	def on_sclOSDTimeout_format_value(self, scale, value):
		if value > 60.0:
			return _("forever")
		if value < 1:
			return "%sms" % int(value * 1000)
		return "%ss" % value



from scc.constants import SCButtons

BUTTON_ORDER = (
	SCButtons.A,
	SCButtons.B,
	SCButtons.X,
	SCButtons.Y,
	SCButtons.BACK,
	SCButtons.C,
	SCButtons.START,
	SCButtons.LB,
	SCButtons.RB,
	SCButtons.LT,
	SCButtons.RT,
	SCButtons.STICKPRESS,
	SCButtons.LPAD,
	SCButtons.RPAD,
	SCButtons.RGRIP,
	SCButtons.LGRIP,
)



#!/usr/bin/env python3
"""SC-Controller - Global Settings

Currently setups only one thing...
"""

import logging
import os

from gi.repository import GdkPixbuf, GLib

from scc.gui.editor import ComboSetter, Editor
from scc.gui.userdata_manager import UserDataManager
from scc.paths import get_controller_icons_path, get_default_controller_icons_path
from scc.tools import _

log = logging.getLogger("GS")


class ControllerSettings(Editor, UserDataManager, ComboSetter):
	GLADE = "controller_settings.glade"

	def __init__(self, app, controller, profile_switcher=None):
		UserDataManager.__init__(self)
		self.app = app
		self.controller = controller
		self.profile_switcher = profile_switcher
		self.setup_widgets()
		self.load_icons()
		self._timer = None
		self.app.config.reload()
		self.load_settings()
		self._eh_ids = ()

	def load_icons(self):
		paths = [get_default_controller_icons_path(), get_controller_icons_path()]
		self.load_user_data(paths, "*.svg", None, self.on_icons_loaded)

	def on_icons_loaded(self, icons):
		lstIcons = self.builder.get_object("lstIcons")
		cbIcon = self.builder.get_object("cbIcon")
		paths = [f.get_path() for f in icons]
		for path in sorted(paths):
			filename = os.path.split(path)[-1]
			name = ".".join(filename.split(".")[0:-1])
			if self.controller.get_type() not in name:
				# Ignore images for other types
				continue
			try:
				pb = GdkPixbuf.Pixbuf.new_from_file(path)
			except:
				# Failed to load image
				continue
			lstIcons.append((path, filename, name, pb))

		cfg = self.app.config.get_controller_config(self.controller.get_id())

		if "icon" in cfg:
			# Should be always
			self.set_cb(cbIcon, cfg["icon"], 1)

	def on_Dialog_destroy(self, *a):
		for x in self._eh_ids:
			self.app.dm.disconnect(x)
		self._eh_ids = ()

	def on_btClearControlWith_clicked(self, *a):
		self.builder.get_object("cbControlWith").set_active(0)

	def on_btClearConfirmWith_clicked(self, *a):
		self.builder.get_object("cbConfirmWith").set_active(0)

	def on_btClearCancelWith_clicked(self, *a):
		self.builder.get_object("cbCancelWith").set_active(1)

	def on_exTouchpadRotation_activate(self, ex, *a):
		rvTouchpadRotation = self.builder.get_object("rvTouchpadRotation")
		rvTouchpadRotation.set_reveal_child(not ex.get_expanded())

	def on_exMenuButtons_activate(self, ex, *a):
		rvMenuButtons = self.builder.get_object("rvMenuButtons")
		rvMenuButtons.set_reveal_child(not ex.get_expanded())

	def on_btClearLeftRotation_clicked(self, *a):
		sclLeftRotation = self.builder.get_object("sclLeftRotation")
		sclLeftRotation.set_value(20)

	def on_btClearRightRotation_clicked(self, *a):
		sclRightRotation = self.builder.get_object("sclRightRotation")
		sclRightRotation.set_value(-20)

	def on_rotation_value_changed(self, *a):
		if self._recursing:
			return
		self.save_config()

	def load_settings(self):
		txName = self.builder.get_object("txName")
		sclLED = self.builder.get_object("sclLED")
		cbAlignOSD = self.builder.get_object("cbAlignOSD")
		sclIdleTimeout = self.builder.get_object("sclIdleTimeout")
		sclLeftRotation = self.builder.get_object("sclLeftRotation")
		sclRightRotation = self.builder.get_object("sclRightRotation")
		cbControlWith = self.builder.get_object("cbControlWith")
		cbConfirmWith = self.builder.get_object("cbConfirmWith")
		cbCancelWith = self.builder.get_object("cbCancelWith")

		cfg = self.app.config.get_controller_config(self.controller.get_id())

		self._recursing = True
		txName.set_text(cfg["name"] or "")
		sclLED.set_value(float(cfg["led_level"]))
		sclIdleTimeout.set_value(float(cfg["idle_timeout"]))
		sclLeftRotation.set_value(float(cfg["input_rotation_l"]))
		sclRightRotation.set_value(float(cfg["input_rotation_r"]))
		cbAlignOSD.set_active(cfg["osd_alignment"] != 0)
		self.set_cb(cbControlWith, cfg["menu_control"], keyindex=1)
		self.set_cb(cbConfirmWith, cfg["menu_confirm"], keyindex=1)
		self.set_cb(cbCancelWith, cfg["menu_cancel"], keyindex=1)
		cbConfirmWith.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) == "-")
		cbCancelWith.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) == "-")
		self._recursing = False

	def save_config(self, *a):
		"""Transfers settings from UI back to config"""
		if self._recursing:
			return
		# Get widgets
		txName = self.builder.get_object("txName")
		sclLED = self.builder.get_object("sclLED")
		cbIcon = self.builder.get_object("cbIcon")
		cbAlignOSD = self.builder.get_object("cbAlignOSD")
		sclIdleTimeout = self.builder.get_object("sclIdleTimeout")
		sclLeftRotation = self.builder.get_object("sclLeftRotation")
		sclRightRotation = self.builder.get_object("sclRightRotation")
		cbControlWith = self.builder.get_object("cbControlWith")
		cbConfirmWith = self.builder.get_object("cbConfirmWith")
		cbCancelWith = self.builder.get_object("cbCancelWith")

		# Store data
		cfg = self.app.config.get_controller_config(self.controller.get_id())
		cfg["name"] = txName.get_text()
		cfg["led_level"] = sclLED.get_value()
		cfg["osd_alignment"] = 1 if cbAlignOSD.get_active() else 0
		cfg["idle_timeout"] = sclIdleTimeout.get_value()
		cfg["input_rotation_l"] = sclLeftRotation.get_value()
		cfg["input_rotation_r"] = sclRightRotation.get_value()
		cfg["menu_control"] = cbControlWith.get_model().get_value(cbControlWith.get_active_iter(), 1)
		cfg["menu_confirm"] = cbConfirmWith.get_model().get_value(cbConfirmWith.get_active_iter(), 1)
		cfg["menu_cancel"] = cbCancelWith.get_model().get_value(cbCancelWith.get_active_iter(), 1)

		try:
			cfg["icon"] = cbIcon.get_model().get_value(cbIcon.get_active_iter(), 1)
			if self.profile_switcher:
				self.profile_switcher.update_icon()
		except:
			# Just in case there are no icons at all
			pass

		# Save (almost)
		self.schedule_save_config()

	def schedule_save_config(self, *a):
		"""Schedules config saving in 1s.
		Done to prevent literal madness when user moves slider.
		"""

		def cb(*a):
			self._timer = None
			self.app.save_config()

		if self._timer is not None:
			GLib.source_remove(self._timer)
		self._timer = GLib.timeout_add_seconds(1, cb)

	def on_sclIdleTimeout_format_value(self, scale, value):
		if value <= 180:  # 2 minutes
			return _("%s seconds") % int(value)
		if value % 60 == 0:
			return _("%s minutes") % int(value / 60)
		return _("%sm %ss") % (int(value / 60), int(value % 60))

	def on_sclLED_value_changed(self, scale, *a):
		if self._recursing:
			return
		cfg = self.app.config.get_controller_config(self.controller.get_id())
		cfg["led_level"] = scale.get_value()
		try:
			self.controller.set_led_level(scale.get_value())
		except IndexError:
			# Happens when there is no controller connected to daemon
			pass
		self.schedule_save_config()

	def on_sclIdleTimeout_value_changed(self, scale, *a):
		if self._recursing:
			return
		cfg = self.app.config.get_controller_config(self.controller.get_id())
		cfg["idle_timeout"] = scale.get_value()
		self.schedule_save_config()



"""SC-Controller - Controller Registration.

Dialog that asks a lot of questions to create configuration node in config file.
Most "interesting" thing here may be that this works 100% independently from
daemon.
"""

import json
import logging
import os
import re

import evdev
from gi.repository import GdkPixbuf, GLib, Gtk

from scc.config import Config
from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN, SCButtons
from scc.gui import BUTTON_ORDER
from scc.gui.app import App
from scc.gui.controller_image import ControllerImage
from scc.gui.creg.constants import (
	AXIS_ORDER,
	SDL_AXES,
	SDL_DPAD,
	SDL_TO_SCC_NAMES,
	STICK_PAD_AREAS,
	TRIGGER_AREAS,
)
from scc.gui.creg.data import AxisData, DPadEmuData
from scc.gui.creg.grabs import InputGrabber, StickGrabber, TriggerGrabber
from scc.gui.creg.tester import Tester
from scc.gui.editor import Editor
from scc.paths import get_config_path, get_share_path
from scc.tools import _, clamp, nameof

log = logging.getLogger("CRegistration")


class ControllerRegistration(Editor):
	GLADE = "creg.glade"
	UNASSIGNED_COLOR = "#FFFF0000"  # ARGB
	OBSERVE_COLORS = (
		App.OBSERVE_COLOR,
		# Following just replaces 'full alpha' in ARGB with various alpha values
		App.OBSERVE_COLOR.replace("#FF", "#DF"),
		App.OBSERVE_COLOR.replace("#FF", "#BF"),
		App.OBSERVE_COLOR.replace("#FF", "#9F"),
		App.OBSERVE_COLOR.replace("#FF", "#7F"),
	)

	def __init__(self, app):
		Editor.__init__(self)
		self.app = app
		self._gamepad_icon = GdkPixbuf.Pixbuf.new_from_file(
			os.path.join(self.app.imagepath, "controller-icons", "evdev-0.svg"),
		)
		self._other_icon = GdkPixbuf.Pixbuf.new_from_file(
			os.path.join(self.app.imagepath, "controller-icons", "unknown.svg"),
		)
		self._axis_data = [AxisData(name, xy) for (name, xy) in AXIS_ORDER]
		self.setup_widgets()
		self._controller_image = None
		self._evdevice = None
		self._tester = None
		self._grabber = None
		self._input_axes = {}
		self._mappings = {}
		self._hilights = {}
		self._unassigned = set()
		self._hilighted_area = None
		self.refresh_devices()

	def setup_widgets(self):
		Editor.setup_widgets(self)
		cursors = {}
		for axis in self._axis_data:
			if "trig" in axis.name:
				continue
			axis.cursor = cursors[axis.area] = cursors.get(axis.area) or Gtk.Image.new_from_file(
				os.path.join(self.app.imagepath, "test-cursor.svg"),
			)
			axis.cursor.position = [0, 0]
		self.builder.get_object("cbInvert_1").set_active(True)
		self.builder.get_object("cbInvert_3").set_active(True)
		self.builder.get_object("cbInvert_5").set_active(True)

	@staticmethod
	def does_he_looks_like_a_gamepad(dev):
		"""Examines device capabilities and decides if it passes for gamepad.
		Device is considered gamepad-like if has at least one button with
		keycode in gamepad range and at least two axes.
		"""
		# Strip special symbols
		name = re.sub(r"[^a-zA-Z0-9.]+", " ", dev.name.lower())

		# ... but some cheating first
		if "keyboard" in name:
			return False
		if "touchpad" in name:
			return False
		if "mouse" in name:
			return False
		if "gamepad" in name:
			return True
		caps = dev.capabilities(verbose=False)
		if evdev.ecodes.EV_ABS in caps:  # Has axes
			if evdev.ecodes.EV_KEY in caps:  # Has buttons
				for button in caps[evdev.ecodes.EV_KEY]:
					if button >= evdev.ecodes.BTN_0 and button <= evdev.ecodes.BTN_GEAR_UP:
						return True
		return False

	def load_sdl_mappings(self):
		"""Attempts to load mappings from gamecontrollerdb.txt.

		Return True on success.
		"""
		# Build list of button and axes
		buttons = self._tester.buttons
		axes = self._tester.axes

		# Generate database ID
		wordswap = lambda i: ((i & 0xFF) << 8) | ((i & 0xFF00) >> 8)
		# TODO: version?
		weird_id = "%.4x%.8x%.8x%.8x0000" % (
			wordswap(self._evdevice.info.bustype),
			wordswap(self._evdevice.info.vendor),
			wordswap(self._evdevice.info.product),
			wordswap(self._evdevice.info.version),
		)

		# Search in database
		try:
			db = open(os.path.join(get_share_path(), "gamecontrollerdb.txt"))
		except Exception as e:
			log.error("Failed to load gamecontrollerdb")
			log.exception(e)
			return False

		with db:
			for line in db:
				if line.startswith(weird_id):
					log.info("Loading mappings for '%s' from gamecontrollerdb", weird_id)
					log.debug("Buttons: %s", buttons)
					log.debug("Axes: %s", axes)
					for token in line.strip().split(","):
						if ":" in token:
							k, v = token.split(":", 1)
							k = SDL_TO_SCC_NAMES.get(k, k)
							if v.startswith("b") and hasattr(SCButtons, k.upper()):
								try:
									keycode = buttons[int(v.strip("b"))]
								except IndexError:
									log.warning("Skipping unknown gamecontrollerdb button->button mapping: '%s'", v)
									continue
								button = getattr(SCButtons, k.upper())
								self._mappings[keycode] = button
							elif v.startswith("b") and k in SDL_AXES:
								try:
									keycode = buttons[int(v.strip("b"))]
								except IndexError:
									log.warning("Skipping unknown gamecontrollerdb button->axis mapping: '%s'", v)
									continue
								log.info("Adding button -> axis mapping for %s", k)
								self._mappings[keycode] = self._axis_data[SDL_AXES.index(k)]
								self._mappings[keycode].min = STICK_PAD_MIN
								self._mappings[keycode].max = STICK_PAD_MAX
							elif v.startswith("h") and 16 in axes and 17 in axes:
								# Special case for evdev hatswitch
								if v == "h0.1" and k == "dpup":
									self._mappings[16] = self._axis_data[SDL_AXES.index("dpadx")]
									self._mappings[17] = self._axis_data[SDL_AXES.index("dpady")]
							elif k in SDL_AXES:
								try:
									code = axes[int(v.strip("a"))]
								except IndexError:
									log.warning("Skipping unknown gamecontrollerdb axis: '%s'", v)
									continue
								self._mappings[code] = self._axis_data[SDL_AXES.index(k)]
							elif k in SDL_DPAD and v.startswith("b"):
								try:
									keycode = buttons[int(v.strip("b"))]
								except IndexError:
									log.warning("Skipping unknown gamecontrollerdb button->dpad mapping: %s", v)
									continue
								index, positive = SDL_DPAD[k]
								data = DPadEmuData(self._axis_data[index], positive)
								self._mappings[keycode] = data
							elif k == "platform":
								# Not interesting
								pass
							else:
								log.warning("Skipping unknown gamecontrollerdb mapping %s:%s", k, v)
					return True
			log.debug("Mappings for '%s' not found in gamecontrollerdb", weird_id)

		return False

	def generate_mappings(self):
		"""Generates initial mappings, just to have some preset to show.
		"""
		buttons = list(BUTTON_ORDER)
		axes = list(self._axis_data)
		log.info("Generating default mappings")
		log.debug("Buttons: %s", self._tester.buttons)
		log.debug("Axes: %s", self._tester.axes)
		for axis in self._tester.axes:
			self._mappings[axis], axes = axes[0], axes[1:]
			if len(axes) == 0:
				break
		for button in self._tester.buttons:
			self._mappings[button], buttons = buttons[0], buttons[1:]
			if len(buttons) == 0:
				break

	def generate_unassigned(self):
		unassigned = set()
		unassigned.clear()
		assigned_axes = set([x for x in self._mappings.values() if isinstance(x, AxisData)])
		assigned_axes.update([x.axis_data for x in self._mappings.values() if isinstance(x, DPadEmuData)])
		assigned_buttons = set([x for x in self._mappings.values() if x in SCButtons.__members__.values()])
		assigned_buttons.update([x.button for x in self._mappings.values() if isinstance(x, DPadEmuData)])
		for a in BUTTON_ORDER:
			if a not in assigned_buttons:
				if a not in (SCButtons.RGRIP, SCButtons.LGRIP):
					# Grips are not colored red as most of controllers doesn't
					# have them anyway
					unassigned.add(nameof(a))
		for a in TRIGGER_AREAS:
			axis = self._axis_data[TRIGGER_AREAS[a]]
			if axis in assigned_axes and a in unassigned:
				unassigned.remove(a)
			elif axis not in assigned_axes:
				unassigned.add(a)
		for a in STICK_PAD_AREAS:
			area_name, axes = STICK_PAD_AREAS[a]
			has_mapping = bool(sum([self._axis_data[index] in assigned_axes for index in axes]))
			if not has_mapping:
				unassigned.add(area_name)

		hilight = unassigned - self._unassigned
		unhilight = self._unassigned - unassigned
		self._unassigned = unassigned
		for a in hilight:
			self.hilight(a, self.UNASSIGNED_COLOR)
		for a in unhilight:
			self.unhilight(a)

	def generate_raw_data(self):
		cbControllerButtons = self.builder.get_object("cbControllerButtons")
		cbControllerType = self.builder.get_object("cbControllerType")
		buffRawData = self.builder.get_object("buffRawData")
		config = dict(
			buttons={},
			axes={},
			dpads={},
		)

		def axis_to_json(axisdata):
			index = self._axis_data.index(axisdata)
			target_axis, xy = AXIS_ORDER[index]
			min, max = axisdata.min, axisdata.max
			if axisdata.invert:
				min, max = max, min

			rv = dict(axis=target_axis, min=min, max=max)
			if target_axis not in ("ltrig", "rtrig"):
				# Deadzone is generated with assumption that all sticks are left
				# in center position before 'Save' is pressed.
				center = axisdata.min + (axisdata.max - axisdata.min) / 2
				deadzone = abs(axisdata.pos - center) * 2 + 2
				if abs(axisdata.max) < 2:
					# DPADs
					deadzone = 0
				rv["deadzone"] = deadzone

			return rv

		for code, target in self._mappings.items():
			if target in SCButtons.__members__.values():
				config["buttons"][code] = nameof(target)
			elif isinstance(target, DPadEmuData):
				config["dpads"][code] = axis_to_json(target.axis_data)
				config["dpads"][code]["positive"] = target.positive
				config["dpads"][code]["button"] = nameof(target.button)
			elif isinstance(target, AxisData):
				config["axes"][code] = axis_to_json(target)

		group = cbControllerButtons.get_model()[cbControllerButtons.get_active()][0]
		controller = cbControllerType.get_model()[cbControllerType.get_active()][0]
		config["gui"] = {"background": controller, "buttons": self._groups[group]}

		buffRawData.set_text(json.dumps(config, sort_keys=True, indent=4, separators=(",", ": ")))

	def load_buttons(self):
		cbControllerButtons = self.builder.get_object("cbControllerButtons")
		self._groups = {}
		model = cbControllerButtons.get_model()
		groups = json.loads(open(os.path.join(self.app.imagepath, "button-images", "groups.json")).read())
		for group in groups:
			images = [
				GdkPixbuf.Pixbuf.new_from_file(os.path.join(self.app.imagepath, "button-images", "%s.svg" % (b,)))
				for b in group["buttons"][0:4]
			]
			model.append([group["key"]] + images)
			self._groups[group["key"]] = group["buttons"]
		cbControllerButtons.set_active(0)

	def save_registration(self):
		self.generate_raw_data()
		buffRawData = self.builder.get_object("buffRawData")
		jsondata = buffRawData.get_text(buffRawData.get_start_iter(), buffRawData.get_end_iter(), True)
		try:
			os.makedirs(os.path.join(get_config_path(), "devices"))
		except:
			pass

		filename = self._evdevice.name.strip().replace("/", "")
		if self._tester.driver == "hid":
			filename = "%.4x:%.4x-%s" % (self._evdevice.info.vendor, self._evdevice.info.product, filename)

		config_file = os.path.join(
			get_config_path(),
			"devices",
			"%s-%s.json"
			% (
				self._tester.driver,
				filename,
			),
		)

		open(config_file, "w").write(jsondata)
		log.debug("Controller configuration '%s' written", config_file)

		self.kill_tester()
		self.window.destroy()
		GLib.timeout_add_seconds(1, self.app.dm.rescan)

	def on_buffRawData_changed(self, buffRawData, *a):
		btNext = self.builder.get_object("btNext")
		jsondata = buffRawData.get_text(buffRawData.get_start_iter(), buffRawData.get_end_iter(), True)
		try:
			json.loads(jsondata)
			btNext.set_sensitive(True)
		except Exception:
			# User can modify generated json code before hitting save,
			# but if he writes something unparsable, save button is disabled
			btNext.set_sensitive(False)

	def on_ibHIDWarning_response(self, *a):
		rvHIDWarning = self.builder.get_object("rvHIDWarning")
		rvHIDWarning.set_reveal_child(False)

	def on_btNext_clicked(self, *a):
		rvController = self.builder.get_object("rvController")
		tvDevices = self.builder.get_object("tvDevices")
		stDialog = self.builder.get_object("stDialog")
		btBack = self.builder.get_object("btBack")
		btNext = self.builder.get_object("btNext")
		pages = stDialog.get_children()
		index = pages.index(stDialog.get_visible_child())
		if index == 0:
			model, iter = tvDevices.get_selection().get_selected()
			dev = evdev.InputDevice(model[iter][0])
			if (dev.info.vendor, dev.info.product) == (0x054C, 0x09CC):
				# Special case for PS4 controller
				cbDS4 = self.builder.get_object("cbDS4")
				imgDS4 = self.builder.get_object("imgDS4")
				imgDS4.set_from_file(os.path.join(self.app.imagepath, "ds4-small.svg"))
				cbDS4.set_active(Config()["drivers"]["ds4drv"])
				stDialog.set_visible_child(pages[3])
				btBack.set_sensitive(True)
				btNext.set_label("_Restart Emulation")
				return
			stDialog.set_visible_child(pages[1])
			self.load_buttons()
			self.refresh_controller_image()
			rvController.set_reveal_child(True)
			self.load_buttons()
			self.refresh_controller_image()
			btBack.set_sensitive(True)
		elif index == 1:
			# Disable Next button and determine which driver should be used
			btNext.set_sensitive(False)
			model, iter = tvDevices.get_selection().get_selected()
			dev = evdev.InputDevice(model[iter][0])
			self.prepare_registration(dev)
		elif index == 2:
			self.save_registration()
		elif index == 3:
			# Next pressed on DS4 info page
			# where it means 'Restart Emulation and close'
			self.app.dm.stop()
			GLib.timeout_add_seconds(1, self.app.dm.start)
			self.kill_tester()
			self.window.destroy()

	def on_btBack_clicked(self, *a):
		stDialog = self.builder.get_object("stDialog")
		btBack = self.builder.get_object("btBack")
		btNext = self.builder.get_object("btNext")
		pages = stDialog.get_children()
		index = pages.index(stDialog.get_visible_child())
		if index == 1:
			stDialog.set_visible_child(pages[0])
			btBack.set_sensitive(False)
			btNext.set_sensitive(True)
		elif index == 2:
			stDialog.set_visible_child(pages[1])
			rvController = self.builder.get_object("rvController")
			self._controller_image.get_parent().remove(self._controller_image)
			rvController.add(self._controller_image)
			btNext.set_label("_Next")
		elif index == 3:
			stDialog.set_visible_child(pages[0])
			btNext.set_label("_Next")
			btBack.set_sensitive(False)
			btNext.set_sensitive(True)

	def on_cbDS4_toggled(self, button):
		config = Config()
		config["drivers"]["ds4drv"] = button.get_active()
		config.save()

	def prepare_registration(self, dev):
		self._evdevice = dev
		self.set_hid_enabled(True)

		def retry_with_evdev(tester, code):
			if tester:
				for s in tester.__signals:
					tester.disconnect(s)
				tester.stop()
			self.set_hid_enabled(False)
			if code != 0:
				log.error("HID driver test failed (code %s)", code)
			if code == 2:
				# Special case, device may be usable, but there is no access
				rvHIDWarning = self.builder.get_object("rvHIDWarning")
				rvHIDWarning.set_reveal_child(False)
			log.debug(f"Trying to use '{dev.path}' with evdev driver...")
			self._tester = Tester("evdev", dev.path)
			self._tester.__signals = [
				self._tester.connect("ready", self.on_registration_ready),
				self._tester.connect("error", self.on_device_open_failed),
			]
			self._tester.start()

		# Check VID+PID info and whether HID driver support is enabled
		if (dev.info.vendor == 0 and dev.info.product == 0) or not self.app.config["drivers"].get("hiddrv"):
			# Not an USB device, skip HID test altogether
			retry_with_evdev(None, 0)
		else:
			log.debug("Trying to use %.4x:%.4x with HID driver...", dev.info.vendor, dev.info.product)
			self._tester = Tester("hid", "%.4x:%.4x" % (dev.info.vendor, dev.info.product))
			self._tester.__signals = [
				self._tester.connect("ready", self.on_registration_ready),
				self._tester.connect("error", retry_with_evdev),
			]
			self._tester.start()

	def on_registration_ready(self, tester):
		cbAccessMode = self.builder.get_object("cbAccessMode")
		fxController = self.builder.get_object("fxController")
		cbEmulateC = self.builder.get_object("cbEmulateC")
		stDialog = self.builder.get_object("stDialog")
		btNext = self.builder.get_object("btNext")

		self.set_cb(cbAccessMode, tester.driver)
		cbAccessMode.set_sensitive(True)

		if not self._mappings:
			self._mappings = {}
			if not self.load_sdl_mappings():
				self.generate_mappings()
			self.generate_unassigned()
			self.generate_raw_data()

		for s in tester.__signals:
			tester.disconnect(s)
		tester.__signals = [
			tester.connect("axis", self.on_tester_axis),
			tester.connect("button", self.on_tester_button),
		]

		self._controller_image.get_parent().remove(self._controller_image)
		fxController.add(self._controller_image)
		pages = stDialog.get_children()
		stDialog.set_visible_child(pages[2])
		cbEmulateC.grab_focus()
		btNext.set_label("_Save")
		btNext.set_sensitive(True)

	def on_device_open_failed(self, *a):
		"""Called when all (or user-selected) driver fails
		to communicate with controller.

		Shoudln't be really possible, but something
		_has_ to happen in such case.
		"""
		d = Gtk.MessageDialog(
			parent=self.window,
			flags=Gtk.DialogFlags.MODAL,
			type=Gtk.MessageType.ERROR,
			buttons=Gtk.ButtonsType.OK,
			message_format=_("Failed to open device"),
		)
		d.run()
		d.destroy()
		self.window.destroy()

	def kill_tester(self, *a):
		"""Called when window is closed"""
		if self._tester:
			tester, self._tester = self._tester, None
			for s in tester.__signals:
				tester.disconnect(s)
			tester.stop()

	def set_hid_enabled(self, enabled):
		"""Enables or disables option to use HID driver"""
		cbAccessMode = self.builder.get_object("cbAccessMode")
		for x in cbAccessMode.get_model():
			if x[0] == "hid":
				x[1] = _("USB HID (recommended)") if enabled else _("USB HID")
				x[2] = enabled

	def on_cbAccessMode_changed(self, cb):
		if self._tester:
			btNext = self.builder.get_object("btNext")
			target = cb.get_model().get_value(cb.get_active_iter(), 0)
			if self._tester.driver != target:
				# User changed driver that should be used, a lot of stuff has
				# to be restarted
				log.debug("User-requested driver change")
				self.kill_tester()
				cb.set_sensitive(False)
				btNext.set_sensitive(False)
				if target == "hid":
					self._tester = Tester(
						"hid", "%.4x:%.4x" % (self._evdevice.info.vendor, self._evdevice.info.product),
					)
				else:
					self._tester = Tester("evdev", self._evdevice.path)
				self._tester.__signals = [
					self._tester.connect("ready", self.on_registration_ready),
					self._tester.connect("error", self.on_device_open_failed),
				]
				GLib.timeout_add_seconds(1, self._tester.start)

	def cbInvert_toggled_cb(self, cb, *a):
		index = int(cb.get_name().split("_")[-1])
		self._axis_data[index].invert = cb.get_active()

	def on_tester_button(self, tester, keycode, pressed):
		if self._grabber:
			return self._grabber.on_button(keycode, pressed)

		what = self._mappings.get(keycode)
		if isinstance(what, AxisData):
			if pressed:
				self.hilight_axis(what, STICK_PAD_MAX)
			else:
				self.hilight_axis(what, STICK_PAD_MIN)
		if isinstance(what, DPadEmuData):
			if pressed:
				self.hilight(nameof(what.button))
				if what.positive:
					self.hilight_axis(what.axis_data, STICK_PAD_MAX)
				else:
					self.hilight_axis(what.axis_data, STICK_PAD_MIN)
			else:
				self.unhilight(nameof(what.button))
				self.hilight_axis(what.axis_data, 0)
		elif what is not None:
			if pressed:
				self.hilight(nameof(what))
			else:
				self.unhilight(nameof(what))

	def on_tester_axis(self, tester, number, value):
		self._input_axes[number] = value
		if self._grabber:
			return self._grabber.on_axis(number, value)

		axis = self._mappings.get(number)
		if axis:
			self.hilight_axis(axis, value)

	def hilight_axis(self, axis, value):
		cursor = axis.cursor
		if cursor is None:
			changed, value = axis.set_position(value)
			value = clamp(STICK_PAD_MIN, value, STICK_PAD_MAX)
			# In this very specific case, trigger uses same min/max as stick
			if value > STICK_PAD_MAX * 2 / 3:
				self.hilight(axis.area, self.OBSERVE_COLORS[0])
			elif value > STICK_PAD_MAX * 1 / 3:
				self.hilight(axis.area, self.OBSERVE_COLORS[1])
			elif value > 0:
				self.hilight(axis.area, self.OBSERVE_COLORS[2])
			elif value > STICK_PAD_MIN * 1 / 3:
				self.hilight(axis.area, self.OBSERVE_COLORS[3])
			elif value > STICK_PAD_MIN * 2 / 3:
				self.hilight(axis.area, self.OBSERVE_COLORS[4])
			else:
				self.unhilight(axis.area)
			# Update raw data if needed
			if changed:
				self.generate_raw_data()
		else:
			parent = cursor.get_parent()
			if parent is None:
				parent = self._controller_image.get_parent()
				parent.add(cursor)
				cursor.show()
			# Make position
			changed, value = axis.set_position(value)
			cursor.position[axis.xy] = clamp(STICK_PAD_MIN, value, STICK_PAD_MAX)
			px, py = cursor.position
			# Grab values
			try:
				trash, ay, trash, ah = self._controller_image.get_area_position(axis.area)
				ax, trash, aw, trash = self._controller_image.get_area_position(axis.area + "TEST")
			except ValueError:
				# Area not found
				cursor.set_visible(False)
				return
			cw = cursor.get_allocation().width
			# Compute center
			x, y = ax + aw * 0.5 - cw * 0.5, ay + aw * 0.5 - cw * 0.5
			# Add pad position
			x += px * aw / STICK_PAD_MAX * 0.5
			y -= py * aw / STICK_PAD_MAX * 0.5
			# Move circle
			parent.move(cursor, x, y)
			cursor.set_visible(True)
			# Update raw data if needed
			if changed:
				self.generate_raw_data()

	def hilight(self, what, color=None):
		self._hilights[what] = color or self.OBSERVE_COLORS[0]
		self._controller_image.hilight(self._hilights)

	def unhilight(self, what):
		if what in self._hilights:
			del self._hilights[what]
		if what in self._unassigned:
			self._hilights[what] = self.UNASSIGNED_COLOR
		self._controller_image.hilight(self._hilights)

	def on_exAdditionalOptions_activate(self, ex):
		rv = self.builder.get_object("rvAdditionalOptions")
		rv.set_reveal_child(not ex.get_expanded())

	def on_exRawData_activate(self, ex):
		rv = self.builder.get_object("rvRawData")
		dialog = self.builder.get_object("Dialog")
		rv.set_reveal_child(not ex.get_expanded())
		if not ex.get_expanded():
			dialog.set_resizable(True)

	def on_area_hover(self, trash, what):
		self.on_area_leave()
		self._hilighted_area = what
		self.hilight(what, self.app.HILIGHT_COLOR)

	def on_area_leave(self, *a):
		if self._hilighted_area:
			self.unhilight(self._hilighted_area)
			self._hilighted_area = None

	def on_area_click(self, trash, what):
		stDialog = self.builder.get_object("stDialog")
		pages = stDialog.get_children()
		index = pages.index(stDialog.get_visible_child())
		if index == 2:
			if what in STICK_PAD_AREAS:
				area_name, axes = STICK_PAD_AREAS[what]
				mnuStick = self.builder.get_object("mnuStick")
				mnuStick._what = "STICKPRESS" if what == "STICK" else what
				mnuStick._axes = [self._axis_data[index] for index in axes]
				mnuStick.popup(None, None, None, None, 1, Gtk.get_current_event_time())
			elif what in TRIGGER_AREAS:
				self._grabber = TriggerGrabber(self, self._axis_data[TRIGGER_AREAS[what]])
			elif hasattr(SCButtons, what):
				self._grabber = InputGrabber(self, getattr(SCButtons, what))

	def on_mnuStickPress_activate(self, *a):
		mnuStick = self.builder.get_object("mnuStick")
		self._grabber = InputGrabber(self, getattr(SCButtons, mnuStick._what), text=_("Press stick or button..."))

	def on_mnuStickmove_activate(self, *a):
		mnuStick = self.builder.get_object("mnuStick")
		self._grabber = StickGrabber(self, mnuStick._axes)

	def on_btCancelInput_clicked(self, *a):
		if self._grabber:
			self._grabber.cancel()

	def refresh_devices(self, *a):
		log.debug("Refreshing device list")
		lstDevices = self.builder.get_object("lstDevices")
		cbShowAllDevices = self.builder.get_object("cbShowAllDevices")
		lstDevices.clear()
		for fname in evdev.list_devices():
			dev = evdev.InputDevice(fname)
			is_gamepad = ControllerRegistration.does_he_looks_like_a_gamepad(dev)
			# bustype 3 is BUS_USB, which is the type used for emulated
			# gamepads. phys is blank for BUS_BLUETOOTH devices.
			if dev.info.bustype == 3 and not dev.phys:
				# Skipping over virtual devices so list doesn't show
				# gamepads emulated by SCC
				continue
			if is_gamepad or cbShowAllDevices.get_active():
				lstDevices.append((fname, dev.name, self._gamepad_icon if is_gamepad else self._other_icon))

	def refresh_controller_image(self, *a):
		cbControllerButtons = self.builder.get_object("cbControllerButtons")
		cbControllerType = self.builder.get_object("cbControllerType")
		rvController = self.builder.get_object("rvController")
		group = cbControllerButtons.get_model()[cbControllerButtons.get_active()][0]
		controller = cbControllerType.get_model()[cbControllerType.get_active()][0]
		config = {"gui": {"background": controller, "buttons": self._groups[group]}}

		if self._controller_image:
			self._controller_image.use_config(config)
		else:
			self._controller_image = ControllerImage(self.app)
			self._controller_image.connect("hover", self.on_area_hover)
			self._controller_image.connect("leave", self.on_area_leave)
			self._controller_image.connect("click", self.on_area_click)
			rvController.add(self._controller_image)



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Custom action

Custom Action page in Action Editor window
"""

import logging

from scc.actions import Action
from scc.gui.ae import AEComponent
from scc.gui.parser import GuiActionParser
from scc.tools import _

log = logging.getLogger("AE.Custom")

__all__ = ["CustomActionComponent"]


class CustomActionComponent(AEComponent):
	GLADE = "ae/custom.glade"
	NAME = "custom"
	PRIORITY = -1
	CTXS = Action.AC_ALL

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		self.parser = GuiActionParser()

	def handles(self, mode, action):
		# Custom Action Editor handles all actions
		return isinstance(action, Action)

	def get_button_title(self):
		return _("Custom Action")

	def load(self):
		if self.loaded:
			return
		AEComponent.load(self)
		try:
			txCustomAction = self.builder.get_object("txCustomAction")
			txCustomAction.set_monospace(True)
		except:
			pass

	def set_action(self, mode, action):
		action = self.editor.generate_modifiers(action, from_custom=True)
		tbCustomAction = self.builder.get_object("tbCustomAction")
		tbCustomAction.set_text(action.to_string(True))

	def on_tbCustomAction_changed(self, tbCustomAction, *a):
		"""Converts text from Custom Action text area into action instance and
		sends that instance back to editor.
		"""
		txCustomAction = self.builder.get_object("txCustomAction")
		txt = tbCustomAction.get_text(tbCustomAction.get_start_iter(), tbCustomAction.get_end_iter(), True)
		if len(txt.strip(" \t\r\n")) > 0:
			action = self.parser.restart(txt).parse()
			self.editor.set_action(action, from_custom=True)

	def shown(self):
		self.editor.set_modifiers_enabled(False)

	def hidden(self):
		self.editor.set_modifiers_enabled(True)



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Menu Only Component

Displays page that can edito only MenuAction
"""

import logging

from scc.actions import Action
from scc.gui.ae import AEComponent
from scc.gui.ae.menu_action import MenuActionCofC
from scc.special_actions import MenuAction, PositionModifier
from scc.tools import _

log = logging.getLogger("AE.SA")

__all__ = ["MenuOnlyComponent"]


class MenuOnlyComponent(AEComponent, MenuActionCofC):
	GLADE = "ae/menu_only.glade"
	NAME = "menu_only"
	CTXS = Action.AC_MENU
	PRIORITY = 0

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		MenuActionCofC.__init__(self)
		self._userdata_load_started = False
		self._recursing = False

	def shown(self):
		if not self._userdata_load_started:
			self._userdata_load_started = True
			self.load_menu_list()

	def set_action(self, mode, action):
		if isinstance(action, PositionModifier):
			action = action.action
		if isinstance(action, MenuAction):
			self._current_menu = action.menu_id

	def get_button_title(self):
		return _("Menu")

	def handles(self, mode, action):
		"""Not visible by default"""
		return False



"""SC-Controller - OSD Menu.

Displays border around area.
"""

from __future__ import annotations

import logging

from gi.repository import GdkX11, Gtk

from scc.lib import xwrappers as X
from scc.osd import OSDWindow
from scc.osd.timermanager import TimerManager

log = logging.getLogger("osd.area")


class Area(OSDWindow, TimerManager):
	BORDER_WIDTH = 2

	def __init__(self) -> None:
		OSDWindow.__init__(self, "osd-area")
		TimerManager.__init__(self)
		self.size = (100, 100)
		self.add(Gtk.Fixed())

	def _add_arguments(self) -> None:
		OSDWindow._add_arguments(self)
		self.argparser.add_argument("--width", type=int, metavar="pixels", default=20, help="""area width in pixels""")
		self.argparser.add_argument(
			"--height", type=int, metavar="pixels", default=-20, help="""area height in pixels""",
		)

	def parse_argumets(self, argv) -> bool:
		if not OSDWindow.parse_argumets(self, argv):
			return False
		self.position = (self.position[0] - self.BORDER_WIDTH, self.position[1] - self.BORDER_WIDTH)
		self.size = (self.args.width + 2 * self.BORDER_WIDTH, self.args.height + 2 * self.BORDER_WIDTH)
		return True

	def compute_position(self) -> tuple[int, int]:
		# Overrides compute_position as Area is requested with exact position
		# on X screen.
		return self.position

	def show(self) -> None:
		OSDWindow.show(self)
		self.realize()
		self.resize(*self.size)
		self.make_hole(self.BORDER_WIDTH)

	def update(self, x: int, y: int, width: int, height: int) -> None:
		"""Update area size and position."""
		self.position = x, y
		self.size = max(1, width), max(1, height)  # Size can't be <1 or GTK will crash
		self.move(*self.position)
		self.resize(*self.size)
		self.make_hole(self.BORDER_WIDTH)

	def make_hole(self, border_width: int) -> None:
		"""Use shape extension to create a hole in the window...

		Area needs only border, rest should be transparent.
		"""
		width, height = self.size
		dpy = X.Display(hash(GdkX11.x11_get_default_xdisplay()))  # I have no idea why this works...
		wid = X.XID(self.get_window().get_xid())

		mask = X.create_pixmap(dpy, wid, width, height, 1)
		xgcv = X.c_void_p()
		gc = X.create_gc(dpy, mask, 0, xgcv)

		X.set_foreground(dpy, gc, 1)
		X.fill_rectangle(dpy, mask, gc, 0, 0, width, height)

		X.set_foreground(dpy, gc, 0)
		X.fill_rectangle(dpy, mask, gc, border_width, border_width, width - 2 * border_width, height - 2 * border_width)

		SHAPE_BOUNDING = 0
		SHAPE_SET = 0
		X.shape_combine_mask(dpy, wid, SHAPE_BOUNDING, 0, 0, mask, SHAPE_SET)

		X.free_gc(dpy, gc)
		X.free_pixmap(dpy, mask)



#!/usr/bin/env python3
"""SC-Controller - Gesture-related GUI stuff.
"""


import logging
from collections import deque

from gi.repository import Gdk, Gtk

from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN
from scc.osd import parse_rgba

log = logging.getLogger("Gestures")


class GestureDraw(Gtk.DrawingArea):
	GRID_PAD = 10
	MAX_STEPS = 5
	LINE_ALPHA = 0.3

	def __init__(self, size, detector):
		Gtk.DrawingArea.__init__(self)
		self._size = size
		self._detector = detector
		self._points = deque(maxlen=256)
		self.connect("draw", self.draw)
		self.set_size_request(size, size)
		self.set_colors()

	def set_colors(self, background="000000FF", line="FF00FFFF", grid="7A7A7AFF", hilight="0030AAFF", **a):
		"""Expects colors in RRGGBB, as stored in config file"""
		self.colors = {
			"background": parse_rgba(background),
			"line": parse_rgba(line),
			"grid": parse_rgba(grid),
			"hilight": parse_rgba(hilight),
		}

	def add(self, x, y):
		factor = self._size / float(STICK_PAD_MAX - STICK_PAD_MIN)
		x -= STICK_PAD_MIN
		y = STICK_PAD_MAX - y
		self._points.append((x * factor, y * factor))
		self.queue_draw()

	def draw(self, another_self, cr):
		resolution = self._detector.get_resolution()
		# hilights = [ [0] * resolution for x in xrange(0, resolution) ]

		# Background
		Gdk.cairo_set_source_rgba(cr, self.colors["background"])
		cr.rectangle(0, 0, self._size, self._size)
		cr.fill()

		# Hilighted boxes
		# Iterates over gesture in progress hilighting apripriate boxes,
		# so user can see what's he doing.
		box_width = float(self._size) / float(resolution)
		col = self.colors["hilight"]
		alpha = col.alpha
		alpha_fallout = alpha * 0.5 / self.MAX_STEPS
		step = 0
		for x, y in reversed(self._detector.get_positions()):
			if step > self.MAX_STEPS:
				break
			col.alpha = alpha - alpha_fallout * step
			Gdk.cairo_set_source_rgba(cr, col)
			cr.rectangle(box_width * x, box_width * y, box_width, box_width)
			cr.fill()
			step += 1
		col.alpha = alpha

		# Grid
		Gdk.cairo_set_source_rgba(cr, self.colors["grid"])
		for i in range(1, resolution):
			cr.move_to(i * box_width, self.GRID_PAD)
			cr.line_to(i * box_width, self._size - self.GRID_PAD)
			cr.stroke()
			cr.move_to(self.GRID_PAD, i * box_width)
			cr.line_to(self._size - self.GRID_PAD, i * box_width)
			cr.stroke()

		# Line
		Gdk.cairo_set_source_rgba(cr, self.colors["line"])
		drawing = False
		for x, y in self._points:
			if drawing:
				cr.line_to(x, y)
			else:
				cr.move_to(x, y)
				drawing = True
		if drawing:
			cr.stroke()



#!/usr/bin/env python3
"""SC-Controller - Slave Mapper

Mapper that is hooked to scc-daemon instance through socket instead of
using libusb directly. Relies to Observe or Lock message being sent by client.

Used by on-screen keyboard.
"""

import logging

from scc.constants import CPAD, DPAD, LEFT, RIGHT, RSTICK, STICK, SCButtons
from scc.mapper import Mapper

log = logging.getLogger("SlaveMapper")


class SlaveMapper(Mapper):
	def __init__(self, profile, scheduler, keyboard=b"SCController Keyboard", mouse=None):
		Mapper.__init__(self, profile, scheduler, keyboard, mouse, None)
		self._feedback_cb = None

	def set_controller(self, c):
		"""Sets controller device, used by some (one so far) actions"""
		raise TypeError("SlaveMapper doesn't connect to controller device")

	def get_controller(self):
		"""Returns assigned controller device or None if no controller is set"""
		raise TypeError("SlaveMapper doesn't connect to controller device")

	def set_feedback_callback(self, cb):
		"""Sets callback called to process haptic feedback effects.

		If callback is set, it's called as callback(hapticdata) every time
		when feedback would happen normally.

		Callback is used here instead of signal so this module doesn't
		depends on GLib
		"""
		self._feedback_cb = cb

	def send_feedback(self, hapticdata):
		"""Simply calls self._feedback_cb, if set. See docstring above.
		"""
		if self._feedback_cb:
			self._feedback_cb(hapticdata)

	def handle_event(self, daemon, what, data):
		"""Handles event sent by scc-daemon.
		Without calling this, SlaveMapper basically does nothing.
		"""
		self.old_buttons = self.buttons
		if what == STICK:
			self.profile.stick.whole(self, data[0], data[1], what)
		elif what == RSTICK:
			self.profile.rstick.whole(self, data[0], data[1], what)
		elif what == SCButtons.LT.name:
			self.profile.triggers[LEFT].trigger(self, *data)
		elif what == SCButtons.RT.name:
			self.profile.triggers[RIGHT].trigger(self, *data)
		elif hasattr(SCButtons, what) or what == "STICKPRESS":
			if what == "STICKPRESS":
				x = SCButtons.STICKPRESS
			else:
				x = getattr(SCButtons, what)
			if data[0]:
				# Pressed
				self.buttons = self.buttons | x
				self.profile.buttons[x].button_press(self)
			else:
				self.buttons = self.buttons & ~x
				self.profile.buttons[x].button_release(self)
				if what == "LPADTOUCH":
					self.profile.pads[LEFT].whole(self, 0, 0, LEFT)
				elif what == "RPADTOUCH":
					self.profile.pads[RIGHT].whole(self, 0, 0, RIGHT)
		elif what in (LEFT, RIGHT, CPAD, DPAD):
			# print what, self.profile.pads[what]
			self.profile.pads[what].whole(self, data[0], data[1], what)
		else:
			print(">>>", what, data)
		self.generate_events()



"""DWSNC - Doing Weird Things in Name of Compatibility

This module, when imported, applies various fixes and monkey-patching to allow
application to run with older versions of GLib and/or GTK.
"""

import os

from gi.repository import GObject, Gtk


def fix_label_missing_set_XYalign_methods():
	"""Fix Gtk.Label missing set_xalign and set_yalign methods with older
	versions of Gtk.

	Prevents crashing, but alings are ignored.
	"""
	Gtk.Label.set_xalign = Gtk.Label.set_yalign = lambda *a: None


def child_get_property(parent, child, propname):
	"""Wrapper for child_get_property, which pygobject doesn't properly
	introspect
	"""
	value = GObject.Value()
	value.init(GObject.TYPE_INT)
	parent.child_get_property(child, propname, value)
	return value.get_int()


def headerbar(bar):
	"""Moves all buttons from left to right (and vice versa) if user's desktop
	environment is identified as Unity.

	Removes 'icon' button otherwise
	"""
	bar.set_decoration_layout(":minimize,close")
	# Not outside of Unity


IS_UNITY = False
IS_GNOME = False
IS_KDE = False

if "XDG_CURRENT_DESKTOP" in os.environ:
	if "GNOME" in os.environ["XDG_CURRENT_DESKTOP"].split(":"):
		IS_GNOME = True

	if "KDE" in os.environ["XDG_CURRENT_DESKTOP"].split(":"):
		IS_KDE = True

	if "Unity" in os.environ["XDG_CURRENT_DESKTOP"].split(":"):
		# User runs Unity
		IS_UNITY = True

		def _headerbar(bar):
			children = [] + bar.get_children()
			pack_start = []
			pack_end = []
			for c in children:
				if child_get_property(bar, c, "pack-type") == Gtk.PackType.END:
					bar.remove(c)
					pack_start.append(c)
				else:
					bar.remove(c)
					pack_end.append(c)
			if len(pack_end) > 1:
				c, pack_end = pack_end[0], pack_end[1:]
				pack_end.append(c)
			if (Gtk.get_major_version(), Gtk.get_minor_version()) > (3, 10):
				# Old ubuntu has this in order, new Ubuntu has it reversed
				pack_end = reversed(pack_end)
			for c in pack_start:
				bar.pack_start(c)
			for c in pack_end:
				bar.pack_end(c)

		headerbar = _headerbar

if not hasattr(Gtk.Label, "set_xalign"):
	# GTK is old enough
	fix_label_missing_set_XYalign_methods()



#!/usr/bin/env python3
import argparse

from scc.paths import get_daemon_socket, get_pid_file
from scc.sccdaemon import SCCDaemon
from scc.tools import init_logging


def main() -> None:
	"""SCCDaemon launcher."""
	init_logging()
	parser = argparse.ArgumentParser()
	parser.add_argument("profile", type=str, nargs="*")
	parser.add_argument("command", type=str, choices=["start", "stop", "restart", "debug"])
	parser.add_argument(
		"--alone", action="store_true", help="prevent scc-daemon from launching osd-daemon and autoswitch-daemon",
	)
	parser.add_argument(
		"--once", action="store_true", help="use with 'stop' to send single SIGTERM without waiting for daemon to exit",
	)
	parser.add_argument("--foreground", action="store_true", help="run scc-daemon in foreground")
	daemon = SCCDaemon(get_pid_file(), get_daemon_socket())
	args = parser.parse_args()
	daemon.alone = args.alone

	profile = " ".join(args.profile)
	if profile:
		daemon.set_default_profile(profile)
		# If no default_profile is set, daemon will try to load last used
		# from config

	if args.command == "start":
		daemon.start(foreground=args.foreground)
	elif args.command == "stop":
		daemon.stop(once=args.once)
	elif args.command == "restart":
		daemon.restart()
	elif args.command == "debug":
		daemon.debug()


if __name__ == "__main__":
	main()



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Tilt

Setups DPAD emulation or menu display
"""

import logging

from scc.actions import Action, NoAction, TiltAction
from scc.gui.ae import AEComponent, describe_action
from scc.gui.binding_editor import BindingEditor
from scc.tools import _

log = logging.getLogger("AE.Tilt")

__all__ = ["TiltComponent"]


class TiltComponent(AEComponent, BindingEditor):
	GLADE = "ae/tilt.glade"
	NAME = "tilt"
	CTXS = Action.AC_GYRO
	PRIORITY = 2

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		BindingEditor.__init__(self, app)
		self._recursing = False
		self.actions = [NoAction()] * 6

	def set_action(self, mode, action):
		if isinstance(action, TiltAction):
			self.actions = list(action.actions)
			while len(self.actions) < 6:
				self.actions.append(NoAction())
			self.update_button_desc(action)

	def update_button_desc(self, action):
		for i in range(len(action.actions)):
			self.actions[i] = action.actions[i]
		for i in range(6):
			self.set_button_desc(i)

	def set_button_desc(self, i):
		desc = describe_action(Action.AC_BUTTON, None, self.actions[i])
		l = self.builder.get_object("lblTilt%s" % (i,))
		if l is None:
			l = self.builder.get_object("btTilt%s" % (i,)).get_children()[0]
		l.set_markup(desc)

	def get_button_title(self):
		return _("Tilt")

	def handles(self, mode, action):
		return isinstance(action, TiltAction)

	def update(self):
		self.editor.set_action(TiltAction(*self.actions))

	def on_action_chosen(self, i, action, mark_changed=True):
		self.actions[i] = action
		self.set_button_desc(i)
		self.update()

	def on_btTilt_clicked(self, b):
		"""'Select Tilt Action' handler"""
		i = int(b.get_name())
		action = self.actions[i]
		ae = self.choose_editor(action, "")
		ae.set_title(_("Select Tilt Action"))
		ae.set_input(i, action, mode=Action.AC_BUTTON)
		ae.show(self.editor.window)



"""SC-Controller - OSD.

Common methods for OSD-related stuff
"""

import argparse
import logging
import os
import traceback

import cairo
import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
gi.require_version("GtkLayerShell", "0.1")

from gi.repository import Gdk, GLib, GObject, Gtk

from scc.config import Config
from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN
from scc.controller import Controller
from scc.gui.daemon_manager import DaemonManager
from scc.osd.timermanager import TimerManager
from scc.paths import get_share_path
from scc.tools import set_logging_level

log = logging.getLogger("osd")


class OSDWindow(Gtk.Window):
	# TODO: Get rid of CSS_3_20, maybe just by dropping support
	CSS_3_20 = """
		#osd-menu-item-big-icon, #osd-menu-item-big-icon-selected {
			min-width: 48pt;
			min-height: 48pt;
		}

		#osd-dialog-buttons #osd-menu-item,
		#osd-dialog-buttons #osd-menu-item-selected {
			min-width: 100px;
			margin: 0px 5px 0px 5px;
		}
	"""

	EPILOG = ""
	css_provider = None  # Used by staticmethods

	def __init__(self, wmclass, layer=None) -> None:
		Gtk.Window.__init__(self)
		OSDWindow._apply_css(Config())

		self.argparser = argparse.ArgumentParser(
			description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter, epilog=self.EPILOG,
		)
		self._add_arguments()
		self.exit_code = -1
		self.position = (20, -20)
		self.mainloop = None
		self._controller = None
		self.set_name(wmclass)
		self.set_wmclass(wmclass, wmclass)
		self.using_wlroots = False
		try:
			import gi

			gi.require_version("GtkLayerShell", "0.1")
			from gi.repository import GtkLayerShell

			if GtkLayerShell.is_supported():
				self.using_wlroots = True
				self.x_layer_anchor = GtkLayerShell.Edge.LEFT
				self.y_layer_anchor = GtkLayerShell.Edge.BOTTOM
				GtkLayerShell.init_for_window(self)
				GtkLayerShell.set_layer(self, layer if layer is not None else GtkLayerShell.Layer.TOP)
				GtkLayerShell.set_anchor(self, self.x_layer_anchor, True)
				GtkLayerShell.set_anchor(self, self.y_layer_anchor, True)
				self.layer_shell = GtkLayerShell
		except ImportError:
			pass
		if not self.using_wlroots:
			self.set_decorated(False)
			self.stick()
			self.set_skip_taskbar_hint(True)
			self.set_skip_pager_hint(True)
			self.set_keep_above(True)
			self.set_type_hint(Gdk.WindowTypeHint.NOTIFICATION)

	@staticmethod
	def _apply_css(config: dict) -> None:
		if OSDWindow.css_provider:
			Gtk.StyleContext.remove_provider_for_screen(Gdk.Screen.get_default(), OSDWindow.css_provider)

		colors = {}
		for x in config["osk_colors"]:
			colors["osk_%s" % (x,)] = config["osk_colors"][x]
		for x in config["osd_colors"]:
			colors[x] = config["osd_colors"][x]
		colors = OSDCssMagic(colors)
		try:
			css_file = os.path.join(get_share_path(), "osd-styles", config["osd_style"])
			with open(css_file) as file:
				css = file.read()
			if (Gtk.get_major_version(), Gtk.get_minor_version()) > (3, 20):
				css += OSDWindow.CSS_3_20
			OSDWindow.css_provider = Gtk.CssProvider()
			OSDWindow.css_provider.load_from_data((css % colors).encode("utf-8"))
			Gtk.StyleContext.add_provider_for_screen(
				Gdk.Screen.get_default(), OSDWindow.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER,
			)
		except GLib.Error as e:
			log.error("Failed to apply css with user settings:")
			log.error(e)
			log.error("Retrying with default values")

			OSDWindow.css_provider = Gtk.CssProvider()
			css_file = os.path.join(get_share_path(), "osd-styles", "Classic.gtkstyle.css")
			with open(css_file) as file:
				css = file.read()
			if (Gtk.get_major_version(), Gtk.get_minor_version()) > (3, 20):
				css += OSDWindow.CSS_3_20
			OSDWindow.css_provider.load_from_data((css % colors).encode("utf-8"))
			Gtk.StyleContext.add_provider_for_screen(
				Gdk.Screen.get_default(), OSDWindow.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER,
			)

	def _add_arguments(self) -> None:
		"""Should be overriden AND called by child class."""
		self.argparser.add_argument(
			"-x",
			type=int,
			metavar="pixels",
			default=20,
			help="""horizontal position in pixels, from left side of screen.
			Use negative value to specify as distance from right side (default: 20)""",
		)
		self.argparser.add_argument(
			"-y",
			type=int,
			metavar="pixels",
			default=-20,
			help="""vertical position in pixels, from top side of screen.
			Use negative value to specify as distance from bottom side (default: -20)""",
		)
		self.argparser.add_argument("--controller", type=str, help="""id of controller to use""")
		self.argparser.add_argument("-d", action="store_true", help="""display debug messages""")

	def choose_controller(self, daemonmanager: DaemonManager) -> Controller:
		"""Return first available controller, or, if --controller argument was specified, controller with a matching ID."""
		if self.args.controller:
			self._controller = self.daemon.get_controller(self.args.controller)
		elif self.daemon.has_controller():
			self._controller = self.daemon.get_controllers()[0]
		return self._controller

	def get_controller(self) -> Controller:
		"""Return controller chosen by choose_controller."""
		return self._controller

	def parse_argumets(self, argv) -> bool:
		"""Returns True on success."""
		try:
			self.args = self.argparser.parse_args(argv[1:])
		except SystemExit:
			return False
		except BaseException:  # Includes SystemExit
			log.error(traceback.format_exc())
			return False
		del self.argparser
		self.position = (self.args.x, self.args.y)
		if self.args.d:
			set_logging_level(True, True)
		return True

	def make_window_clicktrough(self):
		(width, height) = self.get_size()
		surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
		surface_ctx = cairo.Context(surface)
		surface_ctx.set_source_rgba(0.0, 0.0, 0.0, 0.0)
		surface_ctx.set_operator(cairo.OPERATOR_SOURCE)
		surface_ctx.paint()
		reg = Gdk.cairo_region_create_from_surface(surface)
		self.input_shape_combine_region(reg)

	def get_active_screen_geometry(self):
		"""Return geometry of active screen or None if active screen cannot be determined."""
		screen = self.get_window().get_screen()
		active_window = screen.get_active_window()
		if active_window:
			monitor = screen.get_monitor_at_window(active_window)
			if monitor is not None:
				return screen.get_monitor_geometry(monitor)
		return None

	def compute_position(self):
		"""Adjust position for currently active screen (display)."""
		x, y = self.position
		width, height = self.get_window_size()
		geometry = self.get_active_screen_geometry()
		if geometry:
			if x < 0:
				x = x + geometry.x + geometry.width - width
			else:
				x = x + geometry.x
			if y < 0:
				y = y + geometry.y + geometry.height - height
			else:
				y = geometry.y + y

		return x, y

	def get_window_size(self):
		return self.get_window().get_width(), self.get_window().get_height()

	def show(self) -> None:
		self.get_children()[0].show_all()
		self.realize()
		self.get_window().set_override_redirect(True)

		x, y = self.compute_position()
		if self.using_wlroots:
			if x < 0:
				self.layer_shell.set_anchor(self, self.x_layer_anchor, False)
				self.x_layer_anchor = self.layer_shell.Edge.RIGHT
				self.layer_shell.set_anchor(self, self.x_layer_anchor, True)
				x = -x
			if y < 0:
				self.layer_shell.set_anchor(self, self.y_layer_anchor, False)
				self.y_layer_anchor = self.layer_shell.Edge.BOTTOM
				self.layer_shell.set_anchor(self, self.y_layer_anchor, True)
				y = -y
			self.layer_shell.set_margin(self, self.x_layer_anchor, x)
			self.layer_shell.set_margin(self, self.y_layer_anchor, y)
		else:  # X11
			if x < 0:  # Negative X position is counted from right border
				x = Gdk.Screen.width() - self.get_allocated_width() + x + 1
			if y < 0:  # Negative Y position is counted from bottom border
				y = Gdk.Screen.height() - self.get_allocated_height() + y + 1
			self.move(x, y)

		Gtk.Window.show(self)
		self.make_window_clicktrough()

	def on_controller_lost(self, *a) -> None:
		log.error("Controller lost")
		self.quit(2)

	def on_daemon_died(self, *a) -> None:
		log.error("Daemon died")
		self.quit(2)

	def on_failed_to_lock(self, error: str) -> None:
		log.error("Failed to lock input: %s", error)
		self.quit(3)

	def get_exit_code(self) -> int:
		return self.exit_code

	def run(self) -> None:
		self.mainloop = GLib.MainLoop()
		self.show()
		self.mainloop.run()

	def quit(self, code: int = -1) -> None:
		self.exit_code = code
		if self.mainloop:
			self.mainloop.quit()
		else:
			self.destroy()


class OSDCssMagic(dict):
	"""Basically, I reinvented templating.

	This is passed to string.format, allowing to use some simple expressions in
	addition to normal %(placeholder)s.

	Supported magic:
		%(background)s			- just color
		%(background+10)s		- color, 10 values brighter
		%(background-10)s		- color, 10 values darker
	"""

	def __init__(self, dict_to_wrap) -> None:
		self._dict = dict_to_wrap

	def __getitem__(self, a) -> str:
		if "+" in a:
			key, number = a.rsplit("+", 1)
			rgba = parse_rgba(self[key])
			number = float(number) / 255.0
			rgba.red = min(1.0, rgba.red + number)
			rgba.green = min(1.0, rgba.green + number)
			rgba.blue = min(1.0, rgba.blue + number)
			return "%s%s%s" % (
				hex(int(rgba.red * 255)).split("x")[-1].zfill(2),
				hex(int(rgba.green * 255)).split("x")[-1].zfill(2),
				hex(int(rgba.blue * 255)).split("x")[-1].zfill(2),
			)
		if "-" in a:
			key, number = a.rsplit("-", 1)
			rgba = parse_rgba(self[key])
			number = float(number) / 255.0
			rgba.red = max(0.0, rgba.red - number)
			rgba.green = max(0.0, rgba.green - number)
			rgba.blue = max(0.0, rgba.blue - number)
			return "%s%s%s" % (
				hex(int(rgba.red * 255)).split("x")[-1].zfill(2),
				hex(int(rgba.green * 255)).split("x")[-1].zfill(2),
				hex(int(rgba.blue * 255)).split("x")[-1].zfill(2),
			)
		return self._dict[a]


class StickController(GObject.GObject, TimerManager):
	"""Simple utility class that gets fed by with position and emits 'direction' signal that can be used as input for menu navigation.

	Signals:
		direction(horisontal, vertical)

		Both values are one of -1, 0, 1 for left/none/right.
	"""

	__gsignals__ = {
		"direction": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
	}
	REPEAT_DELAY = 0.2
	DIRECTION_TO_XY = {
		0: (0, 0),
		4: (1, 0),
		6: (-1, 0),
		2: (0, 1),
		8: (0, -1),
	}

	def __init__(self):
		GObject.GObject.__init__(self)
		TimerManager.__init__(self)
		self._direction = 0

	def _move(self, *a):
		self.emit("direction", *self.DIRECTION_TO_XY[self._direction])
		if self._direction != 0:
			self.timer("move", self.REPEAT_DELAY, self._move)
		else:
			self.cancel_timer("move")

	def set_stick(self, *data):
		direction = 0
		# Y
		if data[1] < STICK_PAD_MIN / 2:
			direction = 2
		elif data[1] > STICK_PAD_MAX / 2:
			direction = 8
		# X
		elif data[0] < STICK_PAD_MIN / 2:
			direction = 4
		elif data[0] > STICK_PAD_MAX / 2:
			direction = 6

		if direction != self._direction:
			self._direction = direction
			self._move()


def parse_rgba(col: str) -> Gdk.RGBA:
	"""Parse color specified by a #RRGGBBAA string.

	'#' and 'AA' is optional.
	"""
	# Because GTK can parse everything but theese :(
	alpha = "FF"
	if not col.startswith("#"):
		col = "#" + col
	if len(col) > 7:
		col, alpha = col[0:7], col[7:]
	rgba = Gdk.RGBA()
	if not rgba.parse(col):
		log.warning("Failed to parse RGBA color: %s", col)
	rgba.alpha = float(int(alpha, 16)) / 255.0
	return rgba



"""SC Controller - Global Settings.

Currently setups only one thing...
"""

from __future__ import annotations

import collections
import logging
import os
import threading
from io import StringIO

from gi.repository import GLib

from scc.foreign.vdf import VDFProfile
from scc.foreign.vdffz import VDFFZProfile
from scc.lib.vdf import parse_vdf
from scc.tools import _, get_profiles_path

log = logging.getLogger("IE.ImportVdf")


class ImportVdf:
	PROFILE_LIST = "config/localconfig.vdf"
	STEAMPATH = "~/.steam/steam/"

	def __init__(self) -> None:
		self._profile = None
		self._lstVdfProfiles = self.builder.get_object("tvVdfProfiles").get_model()
		self._q_games = collections.deque()
		self._q_profiles = collections.deque()
		self._s_games = threading.Semaphore(0)
		self._s_profiles = threading.Semaphore(0)
		self._lock = threading.Lock()
		self.__profile_load_started = False
		self._on_preload_finished = None

	def on_grVdfImport_activated(self, *a) -> None:
		if not self.__profile_load_started:
			self.__profile_load_started = True
			threading.Thread(target=self._load_profiles).start()
			threading.Thread(target=self._load_game_names).start()
			threading.Thread(target=self._load_profile_names).start()
		self.on_tvVdfProfiles_cursor_changed()

	def _load_profiles(self) -> None:
		"""Search for file containing list of profiles and read it.

		This is done in thread, with crazy hope that it will NOT crash GTK in the process.
		"""
		p = os.path.join(os.path.expanduser(self.STEAMPATH), "userdata")
		if os.path.exists(p):
			for user in os.listdir(p):
				profilelist = os.path.join(p, user, self.PROFILE_LIST)
				if os.path.isfile(profilelist):
					self._lock.acquire()
					log.debug("Loading profile list from '%s'", profilelist)
					try:
						self._parse_profile_list(profilelist, user)
					except Exception as e:
						log.exception(e)
					self._lock.release()
		GLib.idle_add(self._load_finished)

	def _parse_profile_list(self, filename: str, userid: str) -> None | int:
		"""Parse localconfig.vdf and load game and profile IDs. That is later decoded into name of game and profile name.

		Called from _load_profiles, in thread. Exceptions are catched and logged from there.
		Calls GLib.idle_add to send loaded data into UI.
		"""
		# VDF file is a ISO-8859-1 encoded file. Not UTF-8
		data = parse_vdf(open(filename, encoding="ISO-8859-1"))
		# Sanity check
		if "UserLocalConfigStore" not in data:
			return None
		if "controller_config" not in data["UserLocalConfigStore"]:
			return None

		# Grab config - currently only grabs SC configs!
		cc = data["UserLocalConfigStore"]["controller_config"][userid]["controller_steamcontroller_gordon"][
			"DEFAULT_FOR_TYPE"
		]
		# Go through all games
		listitems = []
		i = 0
		for gameid in cc:
			# skip templates
			if "selected" not in cc[gameid]:
				continue

			if not self._check_for_app_manifest(gameid):
				continue

			profile_id = cc[gameid]["selected"]
			listitems.append((i, gameid, profile_id, None))
			i += 1
			if len(listitems) > 10:
				GLib.idle_add(self.fill_list, listitems)
				listitems = []

		GLib.idle_add(self.fill_list, listitems)
		return i

	def _check_for_app_manifest(self, gameid: int) -> bool:
		"""Check if an app manifest exists for a game.

		It seems like a better idea to only worry about importing configs for games the user already has installed.
		"""
		sa_path = self._find_steamapps()
		if sa_path is None:
			return False
		filename = os.path.join(sa_path, "appmanifest_%s.acf" % (gameid))
		return os.path.exists(filename)

	def _load_game_names(self) -> None:
		"""Load names for game ids in q_games.

		This is done in thread (not in same thread as _load_profiles),
		because it involves searching for apropriate file and parsing it
		entirely.

		Calls GLib.idle_add to send loaded data into UI.
		"""
		sa_path = self._find_steamapps()
		if sa_path is None:
			return
		while True:
			self._s_games.acquire(True)  # Wait until something is added to the queue
			try:
				index, gameid = self._q_games.popleft()
			except IndexError:
				break
			if gameid.isdigit():
				name = _("Unknown App ID %s") % (gameid)
				filename = os.path.join(sa_path, "appmanifest_%s.acf" % (gameid,))
				self._lock.acquire()
				if os.path.exists(filename):
					try:
						data = parse_vdf(open(filename))
						name = data["AppState"]["name"]
					except Exception as e:
						log.error("Failed to load app manifest for '%s'", gameid)
						log.exception(e)
				else:
					log.warning("Skiping non-existing app manifest '%s'", filename)
				self._lock.release()
			else:
				name = gameid
			GLib.idle_add(self._set_game_name, index, name)

	@staticmethod
	def _find_legacy_bin(path) -> str | None:
		"""Search specified folder for any file ending in '_legacy.bin'.

		and returns full path to first matching file.
		Returns None if path doesn't point to directory or there is
		no such file.
		"""
		if os.path.exists(path):
			for f in os.listdir(path):
				if f.endswith("_legacy.bin"):
					return os.path.join(path, f)
		return None

	def _load_profile_names(self) -> None:
		"""Load names for profiles ids in q_profiles.

		This is same as _load_game_names, but for profiles.
		"""
		steam_apps_dir = self._find_steamapps()
		if steam_apps_dir is None:
			return
		content_path = os.path.join(steam_apps_dir, "workshop/content")
		if not os.path.exists(content_path):
			log.warning("Cannot find '%s'; Cannot import anything without it", content_path)
			return
		while True:
			self._s_profiles.acquire(True)  # Wait until something is added to the queue
			try:
				index, gameid, profile_id = self._q_profiles.popleft()
			except IndexError:
				break
			self._lock.acquire()
			for user in os.listdir(content_path):
				filename = os.path.join(content_path, user, profile_id, "controller_configuration.vdf")
				if not os.path.exists(filename):
					# If there is no 'controller_configuration.vdf', try finding *_legacy.bin
					filename = self._find_legacy_bin(os.path.join(content_path, user, profile_id))
				if not filename or not os.path.exists(filename):
					# If not even *_legacy.bin is found, skip to next user
					continue
				log.info("Reading '%s'", filename)
				try:
					data = parse_vdf(open(filename))
					name = data["controller_mappings"]["title"]
					GLib.idle_add(self._set_profile_name, index, name, filename)
					break
				except Exception as e:
					log.error("Failed to read profile name from '%s'", filename)
					log.exception(e)
			else:
				log.warning("Profile %s for game %s not found.", profile_id, gameid)
				name = _("(not found)")
				GLib.idle_add(self._set_profile_name, index, name, None)
			self._lock.release()

	def _load_finished(self) -> None:
		"""Called in main thread after _load_profiles is finished."""
		self.builder.get_object("rvLoading").set_reveal_child(False)
		self.loading = False
		self._s_games.release()
		self._s_profiles.release()
		if self._on_preload_finished:
			cb, data = self._on_preload_finished
			GLib.idle_add(cb, *data)
			self._on_preload_finished = None

	def _find_steamapps(self) -> str | None:
		"""Return path to SteamApps folder or None if it cannot be found.

		This is done because Steam apparently supports both SteamApps and
		steamapps as name for this folder.
		"""
		for x in ("SteamApps", "steamapps", "Steamapps", "steamApps"):
			path = os.path.join(os.path.expanduser(self.STEAMPATH), x)
			if os.path.exists(path):
				return path
		log.warning("Cannot find SteamApps directory")
		return None

	def _set_game_name(self, index, name: str) -> None:
		self._lstVdfProfiles[index][1] = name

	def _set_profile_name(self, index, name: str, filename: str) -> None:
		self._lstVdfProfiles[index][2] = name
		self._lstVdfProfiles[index][3] = filename

	def fill_list(self, items) -> None:
		"""Add items to profile list. Has to run in main thread, otherwise, GTK will crash."""
		for i in items:
			self._lstVdfProfiles.append(i)
			self._q_games.append((i[0], i[1]))
			self._s_games.release()
			self._q_profiles.append((i[0], i[1], i[2]))
			self._s_profiles.release()

	def on_tvVdfProfiles_cursor_changed(self, *a) -> None:
		"""Called when user selects profile.

		Check if file for that profile is known and if yes, enables next page.
		"""
		tvVdfProfiles = self.builder.get_object("tvVdfProfiles")
		model, iter = tvVdfProfiles.get_selection().get_selected()
		filename = None
		if iter:
			filename = model.get_value(iter, 3)

		self.enable_next(filename is not None, self.import_vdf)

	@staticmethod
	def gen_aset_name(base_name: str, set_name: str) -> str:
		"""Generate name for profile converted from action set."""
		if set_name == "default":
			return base_name
		return "." + base_name + ":" + set_name.lower()

	def on_txName_changed(self, *a) -> None:
		"""Called when text in profile name field is changed.

		Basically enables 'Save' button if name is not empty string.
		"""
		txName = self.builder.get_object("txName")
		lblASetsNotice = self.builder.get_object("lblASetsNotice")
		lblASetList = self.builder.get_object("lblASetList")

		btNext = self.enable_next(True, self.vdf_import_confirmed)
		btNext.set_label("Apply")
		btNext.set_use_stock(True)
		if len(self._profile.action_sets) > 1:
			lblASetsNotice.set_visible(True)
			lblASetList.set_visible(True)
			log.info("Imported profile contains action sets")
			lblASetList.set_text(
				"\n".join(
					[
						self.gen_aset_name(txName.get_text().strip(), x)
						for x in self._profile.action_sets
						if x != "default"
					],
				),
			)
		else:
			lblASetsNotice.set_visible(False)
			lblASetList.set_visible(False)
		btNext.set_sensitive(self.check_name(txName.get_text()))

	def on_preload_finished(self, callback, *data) -> None:
		"""Schedules callback to be called after initial profile list is loaded."""
		self._on_preload_finished = (callback, data)

	def set_vdf_file(self, filename: str) -> None:
		# TODO: Jump directly to page
		tvVdfProfiles = self.builder.get_object("tvVdfProfiles")
		iter = self._lstVdfProfiles.append((-1, _("No game"), _("Dropped profile"), filename))
		tvVdfProfiles.get_selection().select_iter(iter)
		self.window.set_page_complete(self.window.get_nth_page(0), True)
		self.window.set_current_page(1)

	def on_btDump_clicked(self, *a) -> None:
		tvError = self.builder.get_object("tvError")
		swError = self.builder.get_object("swError")
		btDump = self.builder.get_object("btDump")
		tvVdfProfiles = self.builder.get_object("tvVdfProfiles")
		model, iter = tvVdfProfiles.get_selection().get_selected()
		filename = model.get_value(iter, 3)

		dump = StringIO()
		dump.write("\nProfile filename: %s\n" % (filename,))
		dump.write("\nProfile dump:\n")
		try:
			dump.write(open(filename).read())
		except Exception as e:
			dump.write("(failed to write: %s)" % (e,))
		tvError.get_buffer().set_text(dump.getvalue())
		swError.set_visible(True)
		btDump.set_sensitive(False)

	def import_vdf(self, filename: str | None = None) -> None:
		grVdfImportFinished = self.builder.get_object("grVdfImportFinished")
		self.next_page(grVdfImportFinished)

		tvVdfProfiles = self.builder.get_object("tvVdfProfiles")
		lblVdfImportFinished = self.builder.get_object("lblVdfImportFinished")
		lblError = self.builder.get_object("lblError")
		tvError = self.builder.get_object("tvError")
		swError = self.builder.get_object("swError")
		lblName = self.builder.get_object("lblName")
		txName = self.builder.get_object("txName")
		btDump = self.builder.get_object("btDump")

		if filename is None:
			model, iter = tvVdfProfiles.get_selection().get_selected()
			filename = model.get_value(iter, 3)
		if filename.endswith(".vdffz"):
			self._profile = VDFFZProfile()
		else:
			# Best quess
			self._profile = VDFProfile()

		failed = False
		error_log = StringIO()
		self._lock.acquire()
		handler = logging.StreamHandler(error_log)
		logging.getLogger().addHandler(handler)
		swError.set_visible(False)
		lblError.set_visible(False)
		lblName.set_visible(True)
		txName.set_visible(True)
		btDump.set_sensitive(True)

		try:
			self._profile.load(filename)
		except Exception as e:
			log.exception(e)
			lblName.set_visible(False)
			txName.set_visible(False)
			txName.set_text("")
			self._profile = None
			failed = True

		logging.getLogger().removeHandler(handler)
		self._lock.release()

		if failed:
			swError.set_visible(True)
			lblError.set_visible(True)
			btDump.set_sensitive(False)

			lblVdfImportFinished.set_text(_("Import failed"))

			error_log.write("\nProfile filename: %s\n" % (filename,))
			error_log.write("\nProfile dump:\n")
			try:
				error_log.write(open(filename).read())
			except Exception as e:
				error_log.write("(failed to write: %s)" % (e,))

			tvError.get_buffer().set_text(error_log.getvalue())
		else:
			if len(error_log.getvalue()) > 0:
				# Some warnings were displayed
				swError.set_visible(True)
				lblError.set_visible(True)

				lblVdfImportFinished.set_text(_("Profile imported with warnings"))

				tvError.get_buffer().set_text(error_log.getvalue())
				txName.set_text(self._profile.name)
			else:
				lblVdfImportFinished.set_text(_("Profile sucessfully imported"))
				txName.set_text(self._profile.name)
			self.on_txName_changed()

	def vdf_import_confirmed(self, *a) -> None:
		name = self.builder.get_object("txName").get_text().strip()

		if len(self._profile.action_sets) > 1:
			# Update ChangeProfileActions with correct profile names
			for x in self._profile.action_set_switches:
				id = int(x._profile.split(":")[-1])
				target_set = self._profile.action_set_by_id(id)
				x._profile = self.gen_aset_name(name, target_set)

			# Save action set profiles
			for k in self._profile.action_sets:
				if k != "default":
					filename = self.gen_aset_name(name, k) + ".sccprofile"
					path = os.path.join(get_profiles_path(), filename)
					self._profile.action_sets[k].save(path)

		self.app.new_profile(self._profile, name)
		GLib.idle_add(self.window.destroy)



#!/usr/bin/env python3

"""SC Controller - OSD Daemon

Controls stuff displayed as OSD.
"""

import logging
import os
import sys
import time
import traceback

import gi

from scc.tools import set_logging_level

gi.require_version("Gtk", "3.0")
gi.require_version("Rsvg", "2.0")
gi.require_version("GdkX11", "3.0")
from gi.repository import GLib

from scc.config import Config
from scc.gui.daemon_manager import DaemonManager
from scc.osd import OSDWindow
from scc.osd.area import Area
from scc.osd.dialog import Dialog
from scc.osd.gesture_display import GestureDisplay
from scc.osd.grid_menu import GridMenu
from scc.osd.hmenu import HorizontalMenu
from scc.osd.keyboard import Keyboard
from scc.osd.menu import Menu
from scc.osd.message import Message
from scc.osd.quick_menu import QuickMenu
from scc.osd.radial_menu import RadialMenu
from scc.special_actions import OSDAction
from scc.tools import shsplit

log = logging.getLogger("osd.daemon")


class OSDDaemon:
	def __init__(self):
		self.exit_code = -1
		self.mainloop = GLib.MainLoop()
		self.config = None
		# hash_of_colors is used to determine if css needs to be reapplied
		# after configuration change
		self._hash_of_colors = -1
		self._visible_messages = {}
		self._window = None
		self._registered = False
		self._last_profile_change = 0
		self._recent_profiles_undo = None

	def quit(self, code=-1):
		self.exit_code = code
		self.mainloop.quit()

	def get_exit_code(self):
		return self.exit_code

	def on_daemon_reconfigured(self, *a):
		log.debug("Reloading config...")
		self.config.reload()
		self._check_colorconfig_change()

	def on_profile_changed(self, daemon, profile):
		name = os.path.split(profile)[-1]
		if name.endswith(".sccprofile") and not name.startswith("."):
			# Ignore .mod and hidden files
			name = name[0:-11]
			recents = self.config["recent_profiles"]
			if len(recents) and recents[0] == name:
				# Already first in recent list
				return

			if time.time() - self._last_profile_change < 2.0:
				# Profiles are changing too fast, probably because user
				# is using scroll wheel over profile combobox
				if self._recent_profiles_undo:
					recents = [] + self._recent_profiles_undo
			self._last_profile_change = time.time()
			self._recent_profiles_undo = [] + recents

			while name in recents:
				recents.remove(name)
			recents.insert(0, name)
			if len(recents) > self.config["recent_max"]:
				recents = recents[0 : self.config["recent_max"]]
			self.config["recent_profiles"] = recents
			self.config.save()
			log.debug("Updated recent profile list")
			self.clear_messages()

	def on_daemon_died(self, *a):
		log.error("Connection to daemon lost")
		self.quit(2)

	def on_daemon_connected(self, *a):
		def success(*a):
			log.info("Sucessfully registered as scc-osd-daemon")
			self._registered = True

		def failure(why):
			log.error("Failed to registered as scc-osd-daemon: %s", why)
			self.quit(1)

		if not self._registered:
			self.daemon.request("Register: osd", success, failure)

	def on_menu_closed(self, m):
		"""Called after OSD menu is hidden from screen"""
		self._window = None
		if m.get_exit_code() == 0:
			# 0 means that user selected item and confirmed selection
			self.daemon.request(
				f"Selected: {m.get_menuid()} {m.get_selected_item_id()}",
				#'Selected: %s' % ( shjoin([
				# m.get_menuid(), m.get_selected_item_id()
				# ])),
				lambda *a: False,
				lambda *a: False,
			)

	def on_message_closed(self, m):
		hsh = m.hash()
		if hsh in self._visible_messages:
			del self._visible_messages[hsh]

	def on_keyboard_closed(self, *a):
		"""Called after on-screen keyboard is hidden from the screen"""
		self._window = None

	def on_gesture_recognized(self, gd):
		"""Called after on-screen keyboard is hidden from the screen"""
		self._window = None
		if gd.get_exit_code() == 0:
			self.daemon.request("Gestured: %s" % (gd.get_gesture(),), lambda *a: False, lambda *a: False)
		else:
			self.daemon.request("Gestured: x", lambda *a: False, lambda *a: False)

	@staticmethod
	def _is_menu_message(m):
		"""Returns True if m starts with 'OSD: [grid|radial]menu'
		or "OSD: dialog"
		"""
		return (
			m.startswith("OSD: menu")
			or m.startswith("OSD: radialmenu")
			or m.startswith("OSD: quickmenu")
			or m.startswith("OSD: gridmenu")
			or m.startswith("OSD: dialog")
			or m.startswith("OSD: hmenu")
		)

	def on_unknown_message(self, daemon, message):
		if not message.startswith("OSD:"):
			return
		if message.startswith("OSD: message"):
			args = shsplit(message)[1:]
			m = Message()
			m.parse_argumets(args)
			hsh = m.hash()
			if hsh in self._visible_messages:
				self._visible_messages[hsh].extend()
				m.destroy()
			else:
				# TODO: Do this only for default position once changing
				# TODO: is allowed
				if len(self._visible_messages):
					tmp = list(self._visible_messages.values())
					height = tmp[0].get_size().height
					x, y = m.position
					while y in [i.position[1] for i in tmp]:
						y -= height + 5
					m.position = x, y
				m.show()
				self._visible_messages[hsh] = m
				m.connect("destroy", self.on_message_closed)
		elif message.startswith("OSD: keyboard"):
			if self._window:
				log.warning("Another OSD is already visible - refusing to show keyboard")
			else:
				args = shsplit(message)[1:]
				self._window = Keyboard(self.config)
				self._window.connect("destroy", self.on_keyboard_closed)
				self._window.parse_argumets(args)
				self._window.show()
				self._window.use_daemon(self.daemon)
		elif message.startswith("OSD: gesture"):
			if self._window:
				log.warning("Another OSD is already visible - refusing to show keyboard")
			else:
				args = shsplit(message)[1:]
				self._window = GestureDisplay(self.config)
				self._window.parse_argumets(args)
				self._window.use_daemon(self.daemon)
				self._window.show()
				self._window.connect("destroy", self.on_gesture_recognized)
		elif self._is_menu_message(message):
			args = shsplit(message)[1:]
			if self._window:
				log.warning("Another OSD is already visible - refusing to show menu")
			else:
				if message.startswith("OSD: hmenu"):
					self._window = HorizontalMenu()
				elif message.startswith("OSD: radialmenu"):
					self._window = RadialMenu()
				elif message.startswith("OSD: quickmenu"):
					self._window = QuickMenu()
				elif message.startswith("OSD: gridmenu"):
					self._window = GridMenu()
				elif message.startswith("OSD: dialog"):
					self._window = Dialog()
				else:
					self._window = Menu()
				self._window.connect("destroy", self.on_menu_closed)
				self._window.use_config(self.config)
				try:
					if self._window.parse_argumets(args):
						self._window.show()
						self._window.use_daemon(self.daemon)
					else:
						log.error("Failed to show menu")
						self._window = None
				except:
					log.error(traceback.format_exc())
					log.error("Failed to show menu")
					self._window = None
		elif message.startswith("OSD: area"):
			args = shsplit(message)[1:]
			if self._window:
				log.warning("Another OSD is already visible - refusing to show area")
			else:
				args = shsplit(message)[1:]
				self._window = Area()
				self._window.connect("destroy", self.on_keyboard_closed)
				if self._window.parse_argumets(args):
					self._window.show()
				else:
					self._window.quit()
					self._window = None
		elif message.startswith("OSD: clear"):
			# Clears active OSD windows
			self.clear_windows()
		else:
			log.warning("Unknown command from daemon: '%s'", message)

	def clear_windows(self):
		if self._window:
			self._window.quit()
			self._window = None
		self.clear_messages(only_long_lasting=False)

	def clear_messages(self, only_long_lasting=True):
		"""Clears all OSD messages from screen.
		If only_long_lasting is True, which is default behaviour on profile
		change, only messages set to last more than 10s are hidden.
		"""
		to_destroy = [] + list(self._visible_messages.values())
		for m in to_destroy:
			if not only_long_lasting or m.timeout <= 0 or m.timeout > OSDAction.DEFAULT_TIMEOUT * 2:
				m.destroy()

	def _check_colorconfig_change(self):
		"""Checks if OSD color configuration is changed and re-applies CSS
		if needed.
		"""
		h = sum([hash(self.config["osd_colors"][x]) for x in self.config["osd_colors"]])
		h += sum([hash(self.config["osk_colors"][x]) for x in self.config["osk_colors"]])
		h += hash(self.config["osd_style"])
		if self._hash_of_colors != h:
			self._hash_of_colors = h
			OSDWindow._apply_css(self.config)
			if self._window and isinstance(self._window, Keyboard):
				self._window.recolor()
				self._window.update_labels()
				self._window.redraw_background()

	def run(self):
		self.daemon = DaemonManager()
		self.config = Config()
		self._check_colorconfig_change()
		self.daemon.connect("alive", self.on_daemon_connected)
		self.daemon.connect("dead", self.on_daemon_died)
		self.daemon.connect("profile-changed", self.on_profile_changed)
		self.daemon.connect("reconfigured", self.on_daemon_reconfigured)
		self.daemon.connect("unknown-msg", self.on_unknown_message)
		self.mainloop.run()


if __name__ == "__main__":
	from scc.tools import init_logging

	init_logging(suffix=" OSD")
	set_logging_level("debug" in sys.argv, "debug" in sys.argv)

	d = OSDDaemon()
	d.run()
	sys.exit(d.get_exit_code())



"""SC Controller - Background

Changes SVG on the fly and uptates that magnificent image on background with it.
Also supports clicking on areas defined in SVG image.
"""


import logging
import re
from collections import OrderedDict
from copy import deepcopy
from math import cos, sin
from math import pi as PI
from typing import Self
from xml.etree import ElementTree as ET

import gi

gi.require_version("Rsvg", "2.0")
from gi.repository import Gdk, GdkPixbuf, GObject, Gtk, Rsvg

# sys.modules.pop('xml.etree.ElementTree', None)
# sys.modules['_elementtree'] = None
# ET = importlib.import_module('xml.etree.ElementTree')

log = logging.getLogger("Background")
ET.register_namespace("", "http://www.w3.org/2000/svg")


class SVGWidget(Gtk.EventBox):
	FILENAME = "background.svg"
	CACHE_SIZE = 50

	__gsignals__ = {
		# Raised when mouse is over defined area
		"hover": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
		# Raised when mouse leaves all defined areas
		"leave": (GObject.SignalFlags.RUN_FIRST, None, ()),
		# Raised user clicks on defined area
		"click": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
	}

	def __init__(self, filename, init_hilighted=True) -> None:
		Gtk.EventBox.__init__(self)
		self.cache = OrderedDict()
		self.areas = []
		self.current_svg: str | bytes

		self.connect("motion-notify-event", self.on_mouse_moved)
		self.connect("button-press-event", self.on_mouse_click)
		self.set_events(Gdk.EventMask.POINTER_MOTION_MASK | Gdk.EventMask.BUTTON_PRESS_MASK)

		self.size_override = None
		self.image_width = 1
		self.image_height = 1
		self.set_image(filename)
		self.image = Gtk.Image()
		if init_hilighted:
			self.hilight({})
		self.add(self.image)
		self.show_all()

	def set_image(self, filename):
		with open(filename) as file:
			self.current_svg = file.read()
		self.cache = OrderedDict()
		self.areas = []
		self.parse_image()

	def parse_image(self):
		"""Goes trought SVG image, searches for all rects named
		'AREA_SOMETHING' and generates area list from it.
		This area list is later used to determine over which button is mouse
		hovering.
		"""
		tree = ET.fromstring(self.current_svg.encode("utf-8"))
		SVGWidget.find_areas(tree, None, self.areas)
		self.image_width = float(tree.attrib["width"])
		self.image_height = float(tree.attrib["height"])

	def resize(self, width, height):
		"""Overrides image size.
		Doesn't keep aspect ratio and causes cache to be flushed,
		so this may be slow and nasty.
		"""
		self.size_override = width, height
		self.cache = OrderedDict()

	def on_mouse_click(self, trash, event):
		area = self.on_mouse_moved(trash, event)
		if area is not None:
			self.emit("click", area)

	def on_mouse_moved(self, trash, event):
		"""Not actual signal handler, just called from App.
		"""
		x_offset = (self.get_allocation().width - self.image_width) / 2
		x = event.x - x_offset
		y = event.y
		for a in self.areas:
			if a.contains(x, y):
				self.emit("hover", a.name)
				return a.name
		self.emit("leave")
		return None

	def get_area(self, id):
		for a in self.areas:
			if a.name == id:
				return a
		return None

	def get_all_by_prefix(self, prefix):
		"""Searchs for areas using specific prefix.
		For prefix "AREA_", returns self.areas arrray. For anything else,
		re-parses current image and searchs recursivelly for anything that matches, so it
		may be good idea to not call this too often.
		"""
		if prefix == "AREA_":
			return self.areas
		lst = []
		tree = ET.fromstring(self.current_svg.encode("utf-8"))
		SVGWidget.find_areas(tree, None, lst, prefix=prefix)
		return lst

	def get_area_position(self, area_id):
		"""Computes and returns area position on image as (x, y, width, height).
		Raises ValueError if such area is not found.
		"""
		# TODO: Maybe cache this?
		a = self.get_area(area_id)
		if a:
			return a.x, a.y, a.w, a.h
		raise ValueError("Area '%s' not found" % (area_id,))

	@staticmethod
	def find_areas(xml, parent_transform, areas, get_colors=False, prefix="AREA_"):
		"""Recursively searches through XML for anything with ID of 'AREA_SOMETHING'
		"""
		for child in xml:
			child_transform = SVGEditor.matrixmul(
				parent_transform or SVGEditor.IDENTITY, SVGEditor.parse_transform(child),
			)
			if str(child.attrib.get("id")).startswith(prefix):
				# log.debug("Found SVG area %s", child.attrib['id'][5:])
				a = Area(child, child_transform)
				if get_colors:
					a.color = None
					if "style" in child.attrib:
						style = {y[0]: y[1] for y in [x.split(":", 1) for x in child.attrib["style"].split(";")]}
						if "fill" in style:
							a.color = SVGWidget.color_to_float(style["fill"])
				areas.append(a)
			else:
				SVGWidget.find_areas(child, child_transform, areas, get_colors=get_colors, prefix=prefix)

	def get_rect_area(self, element):
		"""Returns x, y, width and height of rect element relative to document root.

		element can be specified by its id.
		"""
		if type(element) is str:
			tree = ET.fromstring(self.current_svg.encode("utf-8"))
			# SVGEditor.update_parents(tree)
			element = SVGEditor.get_element(tree, element)
		width, height = 0, 0
		x, y = SVGEditor.get_translation(element, absolute=True)
		if "width" in element.attrib:
			width = float(element.attrib["width"])
		if "height" in element.attrib:
			height = float(element.attrib["height"])

		return x, y, width, height

	@staticmethod
	def color_to_float(colorstr):
		"""Parses color expressed as RRGGBB (as in config) and returns
		three floats of r, g, b, a (range 0 to 1)
		"""
		b, color = Gdk.Color.parse("#" + colorstr.strip("#"))
		if b:
			return color.red_float, color.green_float, color.blue_float, 1
		return 1, 0, 1, 1  # uggly purple

	def hilight(self, buttons):
		"""Hilights specified button, if same ID is found in svg"""
		cache_id = "|".join(["%s:%s" % (x, buttons[x]) for x in buttons])
		if cache_id not in self.cache:
			# Ok, this is close to madness, but probably better than drawing
			# 200 images by hand;
			if len(buttons) == 0:
				# Quick way out - changes are not needed
				tmp = self.current_svg.encode("utf-8") if type(self.current_svg) is str else self.current_svg
				svg = Rsvg.Handle.new_from_data(tmp)
			else:
				# 1st, parse source as XML
				tree = ET.fromstring(self.current_svg)
				# 2nd, change colors of some elements
				for button in buttons:
					el = SVGEditor.find_by_id(tree, button)
					if el is not None:
						SVGEditor.recolor(el, buttons[button])

				# 3rd, turn it back into XML string......
				xml = ET.tostring(tree)

				# ... and now, parse that as XML again......
				svg = Rsvg.Handle.new_from_data(xml)
			while len(self.cache) >= self.CACHE_SIZE:
				self.cache.popitem(False)
			if self.size_override:
				w, h = self.size_override
				self.cache[cache_id] = svg.get_pixbuf().scale_simple(w, h, GdkPixbuf.InterpType.BILINEAR)
			else:
				self.cache[cache_id] = svg.get_pixbuf()

		self.image.set_from_pixbuf(self.cache[cache_id])

	def get_pixbuf(self):
		"""Returns pixbuf of current image"""
		return self.image.get_pixbuf()

	def edit(self):
		"""Returns new Editor instance bound to this widget"""
		return SVGEditor(self)


class Area:
	SPECIAL_CASES = ("LSTICK", "RSTICK", "DPAD", "ABS", "MOUSE", "MINUSHALF", "PLUSHALF", "KEY")

	""" Basicaly just rectangle with name """

	def __init__(self, element, transform):
		self.name = element.attrib["id"].split("_")[1]
		if self.name in Area.SPECIAL_CASES:
			self.name = "_".join(element.attrib["id"].split("_")[1:3])
		self.x, self.y = SVGEditor.get_translation(transform)
		self.w = float(element.attrib.get("width", 0))
		self.h = float(element.attrib.get("height", 0))

	def contains(self, x, y):
		return x >= self.x and y >= self.y and x <= self.x + self.w and y <= self.y + self.h

	def __str__(self):
		return "<Area %s,%s %sx%s>" % (self.x, self.y, self.w, self.h)


class SVGEditor:
	"""Allows some basic edit operations by parsing SVG into dom tree and doing
	unholly mess on that.

	Constructed by SVGWidget.edit(), updates original SVGWidget when commit()
	is called.
	"""

	RE_PARSE_TRANSFORM = re.compile(r"([a-z]+)\(([-0-9\.,]+)\)(.*)")
	IDENTITY = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))

	def __init__(self, svgw: SVGWidget | str) -> None:
		if type(svgw) is str:
			self._svgw = None
			self._tree = ET.fromstring(svgw)
		# elif type(svgw) == unicode:
		# self._svgw = None
		# self._tree = ET.fromstring(svgw.encode("utf-8"))
		else:
			self._svgw = svgw
			if type(svgw.current_svg) is bytes:
				self._tree = ET.fromstring(svgw.current_svg.decode("utf-8"))
			else:
				self._tree = ET.fromstring(svgw.current_svg)

	def commit(self) -> Self:
		"""Sends modified SVG back to original SVGWidget instance.

		Return self.
		"""
		self._svgw.current_svg = ET.tostring(self._tree)
		self._svgw.cache = OrderedDict()
		self._svgw.hilight({})

		return self

	def to_string(self) -> bytes:
		"""Returns modivied SVG as string"""
		return ET.tostring(self._tree)

	def clone_element(self, id):
		"""Grabs element with specified ID, duplicates it and returns created
		element. Returned element may get invalidated when commit() is called.

		Returns None if element cannot be found
		"""
		# SVGEditor.update_parents(self)
		e = SVGEditor.get_element(self, id)
		if e is not None:
			clone = deepcopy(e)
			parent = SVGEditor.find_parent(self._tree, e)
			if parent is not None:
				parent.append(clone)
			return clone
		return None

	def remove_element(self, e: str | ET.Element | None) -> Self:
		"""Removes element with specified ID, or, if element is passed, remove that element.

		If 'id' is None, does nothing.

		Returns self.
		"""
		if type(e) is str:
			e = SVGEditor.get_element(self, e)
		if e is not None:
			parent = SVGEditor.find_parent(self._tree, e)
			if parent is not None:
				parent.remove(e)
		return self

	def keep(self, *ids) -> Self:
		"""Removes all elements but ones with ID specified.

		Keeps child elements as well.
		"""

		def recursive(element) -> None:
			for child in list(element):
				if (
					child.tag.endswith("metadata")
					or child.tag.endswith("defs")
					or child.tag.endswith("defs")
					or child.tag.endswith("namedview")
				):
					recursive(child)
				elif child.attrib.get("id") not in ids:
					element.remove(child)

		recursive(self._tree)
		return self

	@staticmethod
	def find_parent(tree: ET.Element, element: ET.Element) -> ET.Element | None:
		"""Find an element's parent without relying on custom attributes."""
		for parent in tree.iter():
			for child in parent:
				if child is element:
					return parent
		return None

	@staticmethod
	def get_element(tree, id: str) -> ET.Element | None:
		"""Recursively searches through XML until element with specified ID is found.

		Returns element or None, if there is not any.
		"""
		if isinstance(tree, SVGEditor):
			tree = tree._tree

		return SVGEditor.find_by_id(tree, id)

	@staticmethod
	def find_by_id(tree, id: str) -> ET.Element | None:
		"""Recursively searches through XML until element with specified ID is found.

		Returns element or None, if there is not any.
		"""
		for child in tree:
			if "id" in child.attrib and child.attrib["id"] == id:
				return child
			r = SVGEditor.find_by_id(child, id)
			if r is not None:
				return r
		return None

	@staticmethod
	def find_by_tag(tree, tag) -> ET.Element | None:
		"""Recursively searches through XML until element with specified tag is found.

		Returns element or None, if there is not any.
		"""
		for child in tree:
			if child.tag.endswith(tag):
				return child
			r = SVGEditor.find_by_tag(child, tag)
			if r is not None:
				return r
		return None

	@staticmethod
	def recolor(element, color) -> bool:
		"""Changes background color of element.

		If element is group, descends into first element with fill set.

		Returns True on success, False if element cannot be recolored.
		"""
		if (
			element.tag.endswith("path")
			or element.tag.endswith("rect")
			or element.tag.endswith("circle")
			or element.tag.endswith("ellipse")
			or element.tag.endswith("text")
		):
			if "style" in element.attrib:
				style = {y[0]: y[1] for y in [x.split(":", 1) for x in element.attrib["style"].split(";")]}
				if "fill" in style:
					if len(color.strip("#")) == 8:
						style["fill"] = "#%s" % (color[-6:],)
						alpha = float(int(color.strip("#")[0:2], 16)) / 255.0
						style["fill-opacity"] = style["opacity"] = str(alpha)
					else:
						style["fill"] = color
						style["fill-opacity"] = style["opacity"] = "1"
					element.attrib["style"] = ";".join(["%s:%s" % (x, style[x]) for x in style])
					return True
		elif element.tag.endswith("g"):
			# Group, needs to find RECT, CIRCLE or PATH, whatever comes first
			for child in element:
				SVGEditor.recolor(child, color)
			return True
		return False

	@staticmethod
	def _recolor(tree, s_from, s_to) -> None:
		"""Recursive part of recolor_strokes and recolor_background"""
		for child in tree:
			if "style" in child.attrib:
				if s_from in child.attrib["style"]:
					child.attrib["style"] = child.attrib["style"].replace(s_from, s_to)
			SVGEditor._recolor(child, s_from, s_to)

	def recolor_background(self, change_from, change_to) -> Self:
		"""Recursively travels entire DOM tree and changes every matching
		background color into specified color.

		Returns self.
		"""
		s_from = "fill:#%s" % (change_from,)
		s_to = "fill:#%s" % (change_to,)
		SVGEditor._recolor(self._tree, s_from, s_to)
		return self

	def recolor_strokes(self, change_from, change_to) -> Self:
		"""Recursively travels entire DOM tree and changes every matching
		line (stroke) color into specified color.

		Returns self.
		"""
		s_from = "stroke:#%s" % (change_from,)
		s_to = "stroke:#%s" % (change_to,)
		SVGEditor._recolor(self._tree, s_from, s_to)
		return self

	@staticmethod
	def matrixmul(X, Y, *a):
		if len(a) > 0:
			return SVGEditor.matrixmul(SVGEditor.matrixmul(X, Y), a[0], *a[1:])
		return [[sum(a * b for a, b in zip(x, y)) for y in zip(*Y)] for x in X]

	@staticmethod
	def scale(xml, sx, sy=None) -> None:
		"""Changes element scale.
		Creates or updates 'transform' attribute.
		"""
		sy = sy or sx
		SVGEditor.set_transform(
			xml,
			SVGEditor.matrixmul(
				SVGEditor.parse_transform(xml),
				[[sx, 0.0, 0.0], [0.0, sy, 0.0], [0.0, 0.0, 1.0]],
			),
		)

	@staticmethod
	def rotate(xml, a, x, y) -> None:
		"""Changes element rotation.
		Creates or updates 'transform' attribute.
		"""
		a = a * PI / 180.0
		SVGEditor.set_transform(
			xml,
			SVGEditor.matrixmul(
				SVGEditor.parse_transform(xml),
				[[1.0, 0.0, x], [0.0, 1.0, y], [0.0, 0.0, 1.0]],
				[[cos(a), -sin(a), 0], [sin(a), cos(a), 0], [0.0, 0.0, 1.0]],
				[[1.0, 0.0, -x], [0.0, 1.0, -y], [0.0, 0.0, 1.0]],
			),
		)

	@staticmethod
	def translate(xml, x, y) -> None:
		"""Changes element translation.
		Creates or updates 'transform' attribute.
		"""
		SVGEditor.set_transform(
			xml,
			SVGEditor.matrixmul(
				SVGEditor.parse_transform(xml),
				[[1.0, 0.0, x], [0.0, 1.0, y], [0.0, 0.0, 1.0]],
			),
		)

	@staticmethod
	def set_transform(xml, matrix) -> None:
		"""Sets element transformation matrix
		"""
		xml.attrib["transform"] = "matrix(%s,%s,%s,%s,%s,%s)" % (
			matrix[0][0],
			matrix[1][0],
			matrix[0][1],
			matrix[1][1],
			matrix[0][2],
			matrix[1][2],
		)

	@staticmethod
	def get_translation(elm_or_matrix, absolute=False):
		if isinstance(elm_or_matrix, ET.Element):
			elm = elm_or_matrix
			matrix = SVGEditor.parse_transform(elm)
			parent = elm.find("..")
			while parent is not None:
				matrix = SVGEditor.matrixmul(matrix, SVGEditor.parse_transform(parent))
				parent = parent.find("..")
		else:
			matrix = elm_or_matrix

		return matrix[0][2], matrix[1][2]

	@staticmethod
	def get_size(elm):
		width, height = 1, 1
		if "width" in elm.attrib:
			width = float(elm.attrib["width"])
		if "height" in elm.attrib:
			height = float(elm.attrib["height"])
		return width, height

	@staticmethod
	def parse_transform(xml):
		"""Returns element transform data in transformation matrix,"""
		matrix = SVGEditor.IDENTITY
		if "x" in xml.attrib or "y" in xml.attrib:
			x = float(xml.attrib.get("x", 0.0))
			y = float(xml.attrib.get("y", 0.0))
			# Assuming matrix is identity matrix here
			matrix = ((1.0, 0.0, x), (0.0, 1.0, y), (0.0, 0.0, 1.0))
		if "transform" in xml.attrib:
			transform = xml.attrib["transform"]
			match = SVGEditor.RE_PARSE_TRANSFORM.match(transform.strip())
			while match:
				op, values, transform = match.groups()
				if op == "translate":
					translation = [float(x) for x in values.split(",")[0:2]]
					while len(translation) < 2:
						translation.append(0.0)
					x, y = translation
					matrix = SVGEditor.matrixmul(matrix, ((1.0, 0.0, x), (0.0, 1.0, y), (0.0, 0.0, 1.0)))
				elif op == "rotate":
					rotation = [float(x) for x in values.split(",")[0:3]]
					while len(rotation) < 3:
						rotation.append(0.0)
					a, x, y = rotation
					a = a * PI / 180.0
					matrix = SVGEditor.matrixmul(
						matrix,
						[[1.0, 0.0, x], [0.0, 1.0, y], [0.0, 0.0, 1.0]],
						[[cos(a), -sin(a), 0], [sin(a), cos(a), 0], [0.0, 0.0, 1.0]],
						[[1.0, 0.0, -x], [0.0, 1.0, -y], [0.0, 0.0, 1.0]],
					)
				elif op == "scale":
					scale = tuple([float(x) for x in values.split(",")[0:2]])
					if len(scale) == 1:
						sx, sy = scale[0], scale[0]
					else:
						sx, sy = scale
					matrix = SVGEditor.matrixmul(matrix, ((sx, 0.0, 0.0), (0.0, sy, 0.0), (0.0, 0.0, 1.0)))
				elif op == "matrix":
					m = [float(x) for x in values.split(",")][0:6]
					while len(m) < 6:
						m.append(0.0)
					a, b, c, d, e, f = m
					matrix = SVGEditor.matrixmul(matrix, [[a, c, e], [b, d, f], [0, 0, 1]])

				match = SVGEditor.RE_PARSE_TRANSFORM.match(transform.strip())

		return matrix

	@staticmethod
	def set_text(xml, text) -> None:
		has_valid_children = False
		for child in xml:
			if child.tag.endswith("text") or child.tag.endswith("tspan"):
				has_valid_children = True
				SVGEditor.set_text(child, text)
		if not has_valid_children:
			xml.text = text

	def set_labels(self, labels: dict) -> Self:
		"""Replaces text on every element named LABEL_something with coresponding
		value from 'labels' dict.

		Returns self.
		"""

		def walk(xml):
			for child in xml:
				if "id" in child.attrib:
					if child.attrib["id"].startswith("LABEL_"):
						id = child.attrib["id"][6:]
						if id in labels:
							SVGEditor.set_text(child, labels[id])
				walk(child)

		walk(self._tree)
		return self

	@staticmethod
	def add_element(parent, e, **attributes):
		"""Creates new element as child of specified parent or, if 1st argument
		is ET.Element, adds that element.

		Returns created or passed element.
		"""
		if not isinstance(e, ET.Element):
			attributes = {k: str(attributes[k]) for k in attributes}
			e = ET.Element(e, attributes)
		parent.append(e)
		return e

	@staticmethod
	def load_from_file(filename):
		tree = ET.fromstring(open(filename).read())
		return SVGEditor.find_by_tag(tree, "g")



"""SC Controller - OSD Mode Mapper

Very special case of mapper used when main application is launched in "odd mode".
That means it's drawn in OSD layer, cannot be clicked and cannot react to
keyboard. This mapper emulates input events on it using GTK methods.

Mouse movement (but not buttons) are passed to uinput as usuall.
"""

import logging

from gi.repository import Gdk, GLib, Gtk

from scc.constants import SCButtons
from scc.gui.gdk_to_key import KEY_TO_GDK, KEY_TO_KEYCODE
from scc.osd.slave_mapper import SlaveMapper
from scc.uinput import Keys

log = logging.getLogger("OSDModMapper")


class OSDModeMapper(SlaveMapper):
	def __init__(self, app, profile):
		SlaveMapper.__init__(self, profile, None, keyboard="osd", mouse="osd")
		self.app = app
		self.set_special_actions_handler(self)
		self.target_window = None

	def on_sa_restart(self, *a):
		"""Restart / exit handler"""
		self.app.quit()

	def set_target_window(self, w):
		self.target_window = w

	def create_keyboard(self, name):
		return OSDModeKeyboard(self)

	def create_mouse(self, name):
		return OSDModeMouse(self)


class OSDModeKeyboard:
	"""Emulates uinput keyboard emulator"""

	def __init__(self, mapper):
		self.mapper = mapper
		self.display = Gdk.Display.get_default()
		self.manager = self.display.get_device_manager()
		self.device = [
			x for x in self.manager.list_devices(Gdk.DeviceType.MASTER) if x.get_source() == Gdk.InputSource.KEYBOARD
		][0]

	def pressEvent(self, keys):
		for k in keys:
			event = Gdk.Event.new(Gdk.EventType.KEY_PRESS)
			event.time = Gtk.get_current_event_time()
			event.hardware_keycode = KEY_TO_KEYCODE[k]
			event.keyval = KEY_TO_GDK[k]
			event.window = self.mapper.target_window
			event.set_device(self.device)
			Gtk.main_do_event(event)

	def releaseEvent(self, keys=[]):
		for k in keys:
			event = Gdk.Event.new(Gdk.EventType.KEY_RELEASE)
			event.time = Gtk.get_current_event_time()
			event.hardware_keycode = KEY_TO_KEYCODE[k]
			event.keyval = KEY_TO_GDK[k]
			event.window = self.mapper.target_window
			event.set_device(self.device)
			Gtk.main_do_event(event)


class OSDModeMouse:
	"""Emulates uinput keyboard emulator too"""

	def __init__(self, mapper):
		self.mapper = mapper
		self.display = Gdk.Display.get_default()
		self.manager = self.display.get_device_manager()
		self.device = [
			x for x in self.manager.list_devices(Gdk.DeviceType.MASTER) if x.get_source() == Gdk.InputSource.MOUSE
		][0]

	def synEvent(self, *a):
		pass

	def keyEvent(self, key, val) -> None:
		tp = Gdk.EventType.BUTTON_PRESS if val else Gdk.EventType.BUTTON_RELEASE
		event = Gdk.Event.new(tp)
		event.button = int(key) - Keys.BTN_LEFT + 1
		window, event.x, event.y = Gdk.Window.at_pointer()
		screen, x, y, mask = Gdk.Display.get_default().get_pointer()
		event.x_root, event.y_root = x, y

		gtk_window = None
		for w in Gtk.Window.list_toplevels():
			if w.get_window():
				if window.get_toplevel().get_xid() == w.get_window().get_xid():
					gtk_window = w
					break
		if gtk_window:
			if gtk_window.get_type_hint() == Gdk.WindowTypeHint.COMBO:
				# Special case, clicking on combo does nothing, so
				# pressing "space" is emulated instead.
				if not val:
					return
				event = Gdk.Event.new(Gdk.EventType.KEY_PRESS)
				event.time = Gtk.get_current_event_time()
				event.hardware_keycode = 65
				event.keyval = Gdk.KEY_space
				event.window = self.mapper.target_window
		event.time = Gtk.get_current_event_time()
		event.window = window
		event.set_device(self.device)
		Gtk.main_do_event(event)


class OSDModeMappings:
	ICONS = {
		"imgOsdmodeAct": SCButtons.A,
		"imgOsdmodeClose": SCButtons.B,
		"imgOsdmodeExit": SCButtons.C,
		"imgOsdmodeSave": SCButtons.Y,
		"imgOsdmodeOK": SCButtons.Y,
	}

	MAIN_WINDOW_BUTTONS = {"vbOsdmodeExit", "vbOsdmodeSave"}
	OTHER_WINDOW_BUTTONS = {"vbOsdmodeExit", "vbOsdmodeAct", "vbOsdmodeClose", "vbOsdmodeOK"}

	def __init__(self, app, mapper, window):
		self.app = app
		self.mapper = mapper
		self.window = window
		self.parent = app.window
		self.first_window = None
		GLib.timeout_add(10, self.move_around)
		self.app.window.connect("focus-in-event", self.on_main_window_focus_in_event)
		self.app.window.connect("focus-out-event", self.on_main_window_focus_out_event)
		self.on_main_window_focus_in_event()

	def set_controller(self, c):
		config = c.load_gui_config(self.app.imagepath or {})
		for name in OSDModeMappings.ICONS:
			w = self.app.builder.get_object(name)
			icon, trash = c.get_button_icon(config, OSDModeMappings.ICONS[name])
			w.set_from_file(icon)

	def on_main_window_focus_in_event(self, *a):
		for x in self.OTHER_WINDOW_BUTTONS:
			self.app.builder.get_object(x).set_visible(False)
		for x in self.MAIN_WINDOW_BUTTONS:
			self.app.builder.get_object(x).set_visible(True)

	def on_main_window_focus_out_event(self, *a):
		for x in self.MAIN_WINDOW_BUTTONS:
			self.app.builder.get_object(x).set_visible(False)
		for x in self.OTHER_WINDOW_BUTTONS:
			self.app.builder.get_object(x).set_visible(True)

	def get_target_position(self):
		pos = self.first_window.get_position()
		size = self.first_window.get_geometry()
		my_size = self.window.get_window().get_geometry()
		tx = pos.x + 0.5 * (size.width - my_size.width)
		ty = pos.y + size.height + 100
		return tx, ty

	def show(self):
		self.window.show()
		self.window.get_window().set_override_redirect(True)

	def move_around(self, *a):
		if self.first_window is None:
			active = self.window.get_window().get_screen().get_active_window()
			if active is None:
				return None
			self.first_window = active

		tx, ty = self.get_target_position()
		self.window.get_window().move(tx, ty)
		return True


def direction(x):
	if x >= 1:
		return 1
	if x <= -1:
		return -1
	return 0
